from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    SystemMessage,
    Text,
    ToolCall,
    ToolOutput,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider": "google",
            "model_id": "gemini-2.5-flash",
            "params": {},
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please look up the book with ISBN 0-7653-1178-X and provide detailed info and a recommendation score"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="google_unknown_tool_id",
                            name="get_book_info",
                            args='{"isbn": "0-7653-1178-X"}',
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {"isbn": "0-7653-1178-X"},
                                    "name": "get_book_info",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b'\n\xd1\x08\x01\xd1\xed\x8ao:?\xf6\xc7\xbb.\x96\xa9\xf7\t\xc9\x9bE,\xaf\xd9\xe7X\xf9:\x0c\xc4D\x1bLk?\xea\xce\xdc):e\xfd\xee\x15zM\xcd\x7f\x83\x866\xc9\xd1\xf4T\xcb\xbd\xbd]\xa75\xc7\x11f\xbe\xbb<J\xfd\xa9\xb6)\x0f1v\xc2\x14\xc4\x15\xb5\x96c\x04q\x1e\x8e\xfd\xb2\x9b\x9d\x1ap\xbd)\xa2\xed\xf5\x9b\x14`\x07\n\'\xf1w\xd7\xd6\xb8#Ju\x8f\xc1@)8\xd0f\x1e~,;]\x14\x17i\xa9\xed+\xcdX\xd5%\xb4ufa\x1c\x91~\xe5W\'\xe0\x9f\x96m\x85(\xf0\xad\x18\xf0\xe3K\xd9\xd5OE\x95\xb6\x7f\xd0y\xf3g\xf0\x0c\xcfw\xa3\xdeJQ\xf2\x95\xa2M\xe6\\\xe5\x17i\x95\x14\xb8Y2W\xecr?\x03\x91z\x86\xea\xc12\x88\x04\xb3\xbe\x1aA\\\x97\x05\xb6\xed\xefG\x9f`c\xb5\x02\xa8\xf6K]\xa0\x92Z[\xa2\x00\xf7oi2#\x1b\xe7\x81\x9e\xbc:\xc9W\x9a\xfc\xfa\x8c@\xa4"\x1d3\x0b\xa4\xab>\xfc\x90\xe6\xfd\x90TsW*b"A\xb2`r\x0c\xd3\xccR\xf8\x0cP_\xf4\x88\xf0\xabiT\x048Tk\xec\xf4T\xd7\xd9_\xbf4\x11\x9a\xae\xa1\'R,\xdaS\xc0\xfeQ\xf0B\xa3\x1d\xb70\xca\xc3\x16}"\xcby3\x92 \xbd\x14048\x85\xf0\x91%\xb5 \xc7"K}\xdb\xbe\x102\xdd\x80p\x1aJ\x80\xc3\x04  \xd3\xa5xq\x96\x8a\xbf>\xd9\x0b\x1d\x83\x04\xcc\x97\x83\xb1Jd\x0f\xd0\xe9no\xb1\r.\xfc\xa4\x920pAy\xf6\x08A6\xf5I]^(Nw\x18O\x17\xaf\x11>L0\xb2\xde\xdd\xad\xc9\xd2.\x91g\xc4\x13\xc0\r\xbc\xf7sr\x7f\xe9\x10\xe5\x92\xf2fl\x9b{>d\x03\x95\xc7\xea\x92I!{\xdf\xb5g\xf6\xc4\xfd\x9dv\x1e\xdd\xc4\xae=\x9c$\xd2Mlzm\xb7\x83\x89/\xe0 jx\x01\x8e\x8f\xc6\xd5\xeb\xdc\x92\x08\xe6\xa0\xe00\x82\x86a\x8a\x9dP\xce":\xee\x9f\xd9-<K\xa1\x0c\x8bg\x00\x0e\x02\xe3\x04\xf7\xc5\x14\x85\x06\x00\xd4\x9a)4\x8e\x9aW(\xaaqfn\x08\n\xce\x00SR\xf8m\xd4tL\xd2I|kh\xd4\x16\r\x18"\x92\xc7\xc8\x93qH\x0f\xb6\xfe\xe0\nkao\x01\xfb<\xbd\x01\x1e\xa1!\xfa\xdaJ\xfcaJ\xcc\xdd%E\x01\ne\x01w\xc1\x1a\xd5q\xef/\xb1\xb9\xb2\x99\xa2\xfa\xbf\xd3\x1b\x02\xb6\xef\t5\xbf^\x16Y\xc4\xfc\xc6\x84\x1f\xca\n\x06\xf6<\x8e\x8d\xfeX\xfd\xe89\x98\x12;\xfb\xe7\xa9^\xf4\x82\xe3Q\xe0h\xb3\x8c\x13\x0f\xe93\x06\x11\x1b?`/\x95\xccu\x93\xf50@BH\xed\xdc;\\\xee\x9e\xcd\x03\xd6"A\xcd\xe8\x98+\xa1\xbd\xc3T\xac\xc0\xde\x1080\xd8C\x93i\xff\x99\xe9\xdc\x1c\xe9\x87\xe2\x08\x9f\x05\x12\xcd\xd9\x0c\xd3\x89F\x8dH\x00\xf2R\xbb\xbd\x8f\x7f:\xa5_\xfd>V%4\x12!\xc7u\x93a;\xe8\xf1m\xeaH1\xc7N\x90\xf5f\xa0\x82\x13\xfe\x91R\xdd;y.\xf5!w\x80\x81s\xdb\x9eL\x85\xac\xae\r\xe40\xd3\x1e\xa1\n\x8d\x14|\xd6\xc3\xae\x02\xd6oo\xd7"N\xe6S\nn$IQ\xb7\x9aFM\x83\x80L5\x10\xb0\xb0B\x88?\xc2\xfb\x94`\xb0%\x96JST7\xcf\xd1\xdaD\x1a*\xd0\x93\xafx\xbf\xf7\xd6Qn\x1ezE\xf2\x82G-\xaa\xeel</4\x1d\xa9\xdaF\xf5c\x10\x88\x8c\x8c\xba\xcb>\xf66\xf3\xe8\xb7\xc6\x14A*\x0e\xd5"MDV\xbaj\x88\x96\x8f\xec\x12_m\x05\x97\xe2zH\x14\xb0\xaaE\xc3L\xc91\x11\x12\x87E\xe1w"\x0b\x0ey\xb2\xc6G\xc1N\t){>8\xacz\xdb\xfa&\x05\xf8\xd5\xd1\xe9xH][>\xd0\xfc\x88%E\x19\xc7w\x87\xc8h\xae\xf7\xd8\xfd\xaa\xf7;K\xd4\x91\x13\x039\x87"\x16\xf6D\x93jf\xdc\xee\x93\xab?K\x95!g\xae\xf4\xa2\x0c\xb5;-\xb8\xeb\x89\xfb|\xdahs\xa4\xd8\xcdM;\x8fCf~@r\xf8u+*\xac\xb6\x85(\x17\xde\x1f&\xbc\xde\xd2F\xb0#T\xc6[\xa3{\x13OK\xb4p&\xac\x81\xa8\rE\x84\n\x065\xb17\x81\x8b\x92\x1e{\xb3\xd6\xc7\xc0\x97.2\r@\x12X\xc8\xf1f\x8a\xeb+\x80\x9c\xf5_4Nc=\x1d\xfa\x06\x10L\xe7\x7f&\xebX\x1b\x99\xa5b\xed\xc3\xad\x85\xa5q\xb1\xca\xca\x03W\xfc)L\xf6\xe2\xff\x17V\x07+0\xfe\x15RK\xa4\xdc\xe2\xe7o',
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="google_unknown_tool_id",
                            name="get_book_info",
                            value="Title: Mistborn: The Final Empire, Author: Brandon Sanderson, Pages: 544, Published: 2006-07-25",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"title": "Mistborn: The Final Empire", "publication_year": 2006, "author": "Brandon Sanderson", "pages": 544}'
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "title": "Mistborn: The Final Empire",
                                        "publication_year": 2006,
                                        "author": "Brandon Sanderson",
                                        "pages": 544,
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b"\n\xcb\x04\x01\xd1\xed\x8ao\x1bY\xfa\xcc\xba+HF\xe3aM\x8b\xc7\x87T\xf4F\x84mP8$l\xc9\x16\xdc)\xf6?M}\xd6=&^Y\xae\x9fG\xdd\xaa5\xd5?b\xe6\xf6w*B,\x0ck\xac\x0b\xa8\x1a\t\xcavy|\x89\x97&`\xeb\xdc{x\x12_g\xc2gF\xbe|o\xd5@:fv\xcf9\x86\xb5\xc9\x00|9\x18D\xdb%\xfe\xcf\xc6Np\xa030\xcfV\xa2?x\nB3\x85&\xb62W\x0e\xbc\xf40\x06\x83\xd2J\xe9\x1bGV\xa0\xd1\xb01\xab\xe5\x1co\xe1\x9fq\x03\xee\x13\x9b&ADr\xba\xc3U%w\x9f\xdd\xbd$\x94J\xeb@F\xaf\xc2\x07\xad~/\xcf\xd8\xd3\x9a\xa4,\xf5\x160\xef\x89\x89m\x9c[\xdbJP*\xeaa\xd5\xd9\xbd4bzp\x84\xac\xa1\x87\x04;<\x83\x12\xc4\xd3\xa1\x8f\xd5\x7f\x94\xa1\xf9]\xc8.\x0b;\xbc}{\x87\xc9@h}\x83E\xf2\x9eT\x12\xd5\x7fC\xa8]\xba8\x80e\xfb\\\xe5$q\x9e\rh?W,WH\xc4E\x9d'@\xc9\r\x8b\xde#\xf9Q\xd1lb-\xbfR5 \xf6\xa8\x8e\x04\x9a\\K3\t\xbb\ryg\xfc\xec\xa7]\xcf\xe0b5\x07\xfe\xf8\xa5\\\xb9\x85Z\xe9\xf1\x03\xb1\x18\x05<a\xb8\x99\xb9=4%3\x05\xbc<a\xef\xe1N\x9a\xeb\xff|\x9d\xc4\xb0n\x9b\xf6\x10y'\x06\x01\xa7\x87\x7f\x15R\x15yd1\x08\xaf\xfc=;\xf8V\xe3pr\xe04\x1e\x12=0\x9b\xbb%\xa5\xab\xcf3\xb02\x9b\xf3\xd5w\x90\x9e\xaa\xb6\x84\x82\xaa\xa2\xa8\x9aN\x99\xaa\x0f\xe2\x90\x94C\x95\x1e\xc7=\x01\x0b\xeaN\x0f\xc6fD\xcc\x18\xb21\xeb\xc4FuJG&\xd8kPP\xefegs\xf8\xc2\xc2\x93e\xa1:\x9dY\x11\xc2\xe8\xd4T\t \xdc\xf6h\x8a\x99,N\xcfVE\x1d\x1d\x03p\x03!E\xaa\xd5~\x1f7\xf3\x8bp\xc0\xed\xfc\x03\xc7\xe7\xed\xac\n\xcf;\x15L\xa5\xcb\x15\x1a\xbb\x80\x19k\xfa\xef\x1c\xab\x95h[u\xc3\xaa\x01\xa5\x14\x19\x8a^\xe1\xdfq\x0e\xba\xe2\xe2\x1c\xa5\x0c\xd3\x9e\xcd\xf9cu\xb2-\xbdT\xf6\xb4`\xd5H\xf6\xddc\xc6P\x0f\xf0d\x86\x7f\x93\x8d\xc1#Y[\xaak2\xbe'\xd5\x02\xcbq\xf8<\t\xf3T\xc8;\xe4e\xca_\x8dt\x19\x9f\x88\xa9\xb1\x9c_\xcf\x07\xe0\x9b\xa3l",
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "BookSummary",
                "description": None,
                "schema": {
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"title": "Author", "type": "string"},
                        "pages": {"title": "Pages", "type": "integer"},
                        "publication_year": {
                            "title": "Publication Year",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "pages", "publication_year"],
                    "title": "BookSummary",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [
                {
                    "name": "get_book_info",
                    "description": "Look up book information by ISBN.",
                    "parameters": """\
{
  "properties": {
    "isbn": {
      "title": "Isbn",
      "type": "string"
    }
  },
  "required": [
    "isbn"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider": "google",
            "model_id": "gemini-2.5-flash",
            "params": {},
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please look up the book with ISBN 0-7653-1178-X and provide detailed info and a recommendation score"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="google_unknown_tool_id",
                            name="get_book_info",
                            args='{"isbn": "0-7653-1178-X"}',
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {"isbn": "0-7653-1178-X"},
                                    "name": "get_book_info",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b'\n\xb1\x14\x01\xd1\xed\x8aom<\x01\xa9<\x97\xcc\x03\x0f\xe2+F*\xb1qI\x96~\xb5&9\xcb\x04p\x0e\xc1S+\x03qzx\x96\xf3S\\\x1d\x9f\xde\x1d\xdb\x97\xbf\xa6\xe57K?\x04\xb3\xa9`\x99A\t\xdbs\xfc\xdfU\xbac\xc3\xb3\x8e\xca7\xbf\x9eXF\xac\x08\xf8{\x1cs\x92\xcf\x18e\xe3\xd3r\x88\xdf8\xd2\xc6\xaf\xb0\xc8\xad\x02h\x18cP\xebE\xec\xb2\x81u\xb1\xeb\xa6\x9c$\x95*\xcd7\x87\xabe\x08\xec\xc4K\x02kC#Y\x7f\x06y\x91\x81a\xac\xcd^\xc2\x9b8\xa5\x12\x08\xe7\xece\xe0\xbf\xa6\xe6\x91\x87j\x922)\xdf\xb9\xa8\x14C\xf7M\xd4\xd4o\xa1T\xd8\xc4\x7f\x93c\x1dX\x1e\xbez\xc6\xf4\xbf\x8fD\xe4O\xb7\xbb\xefj\x85\xe5|"\xab\x06n\xac\xb7"t\x17\xa7\xd1_Di<\xb5\xf6\xec\x1b\xe8t\x81\x95\x11\x9a\x7f\x94\xc8XY\x05~5[\xb9v\x14\x04\x87c\xa2l\xf1x@\xa9\x85c\x11\x060\x8a\xce\xdb\xe7\xc9\xf7\xd1mYz8_\x13\xba\x8e\xe7@\x80\x04+u\x1f\x7f\xd3\xbcK\xdb\x10\xe5\xd0\xd18\xbbN\x1c#\x16\xf1CE5T\x81p\xa4:\x07\xba\x8e\xfe\xd6\x0bn\x83:\xfc\x0f\xb1\xc6\xbe\x99\x9f\xdd\xb7\x1e1\xaeB\xa4\x07\xb7&\x16\xebG\xdc\xe4)\x89\xf7\xc4\xfe\xe4\x82\x93\x96\xbd\xf6?Gb\xa4!^V\xf7\xe6c\x10c\xc5>u\x11\xaa\x92Y\'\x1d\x0f\x9f\x97\xcf$\xf5p\xd5c\xf1\x0b8\xa1^@\xfc\xf2\xf1tO\xa5\xdc\x93Huh^\xf7r\xec\x16\xcdc&\x8aok2nC[\xd1\x80\xdb\x9d\xfeg\x1cP\xff\xc5\xe3\xf53\xc2\x94\xea\x9f\x92\xf5\x121L\xa1xy\xbaH\x9c1\xf9\xa5\xa5(\xf7\x1d\xd8\x99\xf5C\xdf\x96.\xf0\x96\xc6\x01\x00H\xd0\x9c:$\x85\xdb\xb1\x7f\x95\x19\x958\xcd\xc9.\xb6\xcb\x05LQ\xa19\x9c\xbb\xab\\,\xa14`\xcc\t\x99\xcd\xa8\xc5\xa6\x0f\x19V"\xd1l7UO\x95\x81\x8c\xfb\x8a\x89\xc3\xe6Z\xb8\xab Tt\x0fj\x99\xd8\x89\xf5tYH\x86? \x18S\xdc\x97\'\xa1\xee\xa3\x0edm\xafC\xe9\xab\xaf\x7f\xe09\x92f{w\xabk\xee\xd9C\xe6y\xd9\xb3\xe9+4\xae\x12\x8c9\x18@*!p\x05\xd3\xf5\xfcJD\xfdS\x10C\x87\x9d(\xe0Z\xd5\x1e\xbc|\xf1\xedW\xf9\xb8\xa8\x8f\x1e(U\xae\xdcr\xea?[A\x11\x82o\x8fR*\x92\x0eP\xa5\x91\xdfg\xda@^\xbd>\x08\xbb}\x06\xd8\xbb\xeeI\x9e\x83\x07\x0c}\xa4\xba\xb0\xe4\x92\x15\x86\x0c\xeaW\x98\xec\xda\xfa\x98\xfb\x9aY\x0c\x87s\x7f\x0ey\x8c\xf6}\x123!_\xbd\r4\xb1\xdd>\x9c\x12\x9ao\xb5H\x8c\x99\xb55P\xf9,`\x02\x02\xe8,Q\x91\xa6\x17\x07\xbc\x12\xa7\x00"\xd6\xa7\xa4-K\x8d5\x19\x01\x9c\x01\xb9\xc2\xcal\xf4t\x8e\xd4\xd8F\xeb\xe9pF!\x90\xb0\xd2E\xb5\x0e1\xdaUV\xebpL\x10\xa4a\x82\xf6Ua\x90\xc1\xf1\x8d"QN\xa9\xb8\x04\xb3\xf3M\x01\x9f<\xde\x0b\xe1\xfe\xb9\xa7M\xd4\x9f7\t\x999\x11W\xb6m\x90\x01\nn\x1e\x16\x9c\x9a\x1e\x15\xd6\x11\x94#ZN\x8fr\x12\xb4\xfeZ\x95\xa9\xad\x9a\xa6\xe4\xf8`\xab!FkC\xee.\x8c\x85\xdc\x99\xbe\x06ye\xc6]\x9d;J\xa8A(y\x00\x07l\xb3(\x08\ny\x013C\r2\xe6s\xde\xd5\x1eG\x01{\xb7\xb4x\x1d\xc4]\x05\xa2 \xb40\xaapD\xf6h\xb0@\xfd\xb38\xb87\x86W\x80V;\x05"\xc0g\xb4\xa9\x13\x8d\x1b\x1c1\x11\xc8\tC\x18s\xc9\xe10\xebg\x07\x0b\xf8\xde\xa3\x7f^`\xc3=J\x0c\x9a\xff\xfa\xab\x04)\xf7\x0cM\x12]4\xa5\x9f\xf3{\xe9mBw\xb5\xbe\x96Pc\xb3\xd0\xc9\x93Z\x92\xb4\xeaE\xc6n<\x164\xc3\x04-\xb6\xb6|\x90r\xaa!e\xb1BCrR\xa0L\x19\x9fI4S\xb7=\xebvD\x8cl\x12\xe0Y\x04\xdb\xdb-\xaf\x9c\x98\xba\xd4\xae\xc9\xb6!\xff\x0e\xd2)\x9e\x03%\xf6\xe4"\xe7\xec\xe3>\x93\xd7\x8d\x87\x82\x91:\xc48\xa6U\xe8\x95\x1a\xd3i\x8c8s1\x00\xad69\x03m\xadX\xea\xfe\xde\xc6&\xf7\xc7\xecrI\xbe\xf7{$R\x04 nV\xbe\x1c\xaa\xbe\x0b\xc6\xf2|\x15B\xc7\xd9\x0e\xad\x81\xbc\xc8\xe0,k\x92\x9f\x16\xfc\xd5\xa8\x95))\x90Df\xbb%w\x9dm\x06%,B\xd60\xc6\xab4\x97\x81\x0e\x9b\xaf\x1e\x101<V\x08\xd8\xa4A\xca\x9dv\xba\nEC2v\xd5#\xe2\xd2\x12\x03\xaf2X\xbaU=C\x0c\x9f\x15\x02\xcb\xf1\xd3\xcc<\x9d\xe7\xcb\xc3\xa4\xfd\x16\x86\'\xa7[\xc8\xe8\xc4(\xd4\xefP\x85\x11_\xaa\xbd\x0b\x1e;\x84\xd4\x0c\xd5\x0b\xae\x94\xf51\x19\x02\xdaZ"\xd4H\x88!\x80\x86 \xf0j\x0b\x92\xf16Y\xa4b\x1a4\xd5\xa6j\x12\xe7\xb0\xd7\x85[\x99\xa3vM]Z\xa7\x10\xb9ru\xab\x0bvx\x8e\x05Iv\xb5 \xc3\xfc\x9ao\xfe\xcf0\xd2\xa2\x9b$\x83]k\x1cd\x8eb\':\xab\x18\x13\xcd\xc7\x8b\x03Pp\xd7\xa0-:\xdf.?\xf3\xcfS\xbes:&\xe1\xa6\xf9\xfb\x90%_\xf0\x17\xcb\xf6\xf1\xd0\xbc\xf8\x8a\xfdO$\xa2\rTv\x90\xafG=8\xdb\xe9\xcaeR\xba<\xf7\x9a\xdc\xc1E\x89\xecH\x19\xe1\xe0\xf3\xd9;\xff\x88\xf2g~;\x97*\x82\xb8\x04\x88\xd5\x9cUJ\xfb6\xf4l\x06\xb3~\n[\x04\xf6!\x8c|\x0c\x83\x9dRy\x99\x19e\xb7\x0b\x88\xe6d\xb6\\{\x03 %\xc0\x1b$\x1e\xb4\xc5z\xee\x10\x10i\xd4P* \xe2F\xdf`\x9av!"\xa8p\xdfY\xfc8Ii\x1d\xcb\xe5\xe5\xb1B\xa7\xcd\xd3\xfb\x90\x8c\xf1\xe0*f\xc5y\xcfrD\x0f\x82S\xd1\xb2\x87\xfdB3<\x87\x8f\x85\x97\xed\xda R\xc7\x81p~\xc1M\xf4\x95\xf6I \xfc=\xd8\xcb>\x92^V\x84\xd1{}@>\xa0N\xb0\xf2\xb1\x88\x7f\x89\x89;\x8a\xff\xd1y\xa5H\xf4\xe2<\xa1\x02Mx\xec\xc4\xe7\x0c\xbe/`\xb0\x1d\xb3\xb1\x9b\x01\xae\xb3C\xe0\xc9\xcc/c\xc1\xae$\xd5\x87\xbf\xf6\xdb<\x15\xec\xf1\xa2!\xf6\x12\x9c\x91\x8f\xb5+\xfd5|\xff]\x11\x80\xa8\x93\xcdy\x13\xc4D$\xec\xd3\x8c\xc9\xeb\xd0\xc6\xa1&\xb5#\xdd\xf0\x99\xba\xa5\x0cj\xa4N\xd3\xc9\xb1\xacc6jT\x04I\xd3\xfd\xf7\x80\x08\xab\xfeq\x01\xfcdr\xba\xa5\x90Z\xf6k\xbb\xb2\xf2\xecs\xdd\xc4l`\xef\xfe\x15\\\xe5~\xa4\x90\xdc\xfc\xf9\xd2\xb5,\x81\xa6\xa6\xbe\x12\xb6\x82C\xa9y\xf3\x8a\xff])\xc1\x0f\xbe\x0e;\xae\xbc\xb0\xaeY\xd9h\xa7q\x11\xfb\xe5,\x8cDDm\x14\xc6\xf1\x85DT\x84\xcb\xc5\xca\xf9\xe2\x91k\x86\x00\xd4\x92`V\xff{7\x195\x81^_-O\xa5\xf3\x01\x8a\x0c\x0c\xb99\xab\xb3N\x11r\x9a\xc1\xa68\xf0\x9dv\xa1u\x90k\x81[d$5\x1e{\xd6H\xbaQ=\x9b\xd3\xe8r\xf9_\xb8(@\x90^\x1b\xcbwl(K\x85\xcf\xc4\xdf\xf4\x9d\xafA^\x8b*{mT\x98\xf1\xeaG\x0b\xe6J\x8b\xa7\x8f\xd9\x97\xd8\xb4\xc3\xae\x10\xc4"\xf7\x14(\xb1*?\x17\xf0\n\x02\x84\xd2\x1e\x86IeB\xa7\x19=B\xad\xa2\xb41WSO\xc435R\xcdi\xa6\x9a]\xc4\x01\xd1T\xbb\x83.\x91\x89WI\x8c3\x89?&e\xe8\xc6\xb0\x82\xc9\xb4\x88\x86fa\xe1$\x11\n\x01d\x8c}\x89(\x87=5u\xff\x89\xac2/5v\x86\xf2\x82\x8cn\xc6\x8b\x10\x16N\x05*.\xd1\xde\xe5\x98\xff\n\x17\xb98\xde\xff\xb2\x9d4.\xc9\xe1\x07\xe7\xcd\xe0\x99X\xe3;\x85\xb4\x98\x8f\xeaQ~\xf6\xc41t\x14\xa6*\x14\xa8\xff\x1b\xa3\xa4\xb5\xafH]\x9a|\x83o\x9b\xaf\xf4\xb7\xe9\x85/x\x0epi\x0b)\xd8\xfb\xd5\xe6\xeb\x99\xe0nV\xf2\xc7\x84\x98\xc9\x14\xd49\xcf\xbc\xb1#\xb0\xb5\xc2o5:\x08\x12\xb9\x90\x9dvp"\xc8*\xccC\xa3\xc9\xe5\x96kF8\x0c\xc7\x1c@\xd5\x13\x89T\xdbz\x84\xbb<o\xb4\x11\xd4Q\x06\x1f\xe1\xb3\xd3\xfb\xf8\xaa\\\x1c\xf8\xac\x93\xc9\xb3z\x9c\xa7\x99&lB\xddx\xe7\x1c|\xb9\xe2\x01fs \xef\x98\xdf\'\xcfy@\xa4\xaa\xe5\x1a\xe5\xa3\x11\xa9L\xe1\xb4\xe5\x07\x91\xad\x11\x0e\xf9E\x8b\r\xc71\xef1\xd4\x0b\xa9\xbftq\x99nH\x12\x82\xc6KFZ\xe8\xc7\xf1\'|\xc5\xe5ie\x13\x19\xfb\x07\xbdV\xe6E\xcb/\xeb&\xec;\xaa\xad\x90+7ba\x9a\xb2\x81j\xf0j\xff\xec<\xf6\xc1\xb9\x06\n\r]A\xa6\xe3\x1d50\xb3\xb2`\xfb\xb1\xf9\xa9,\x16\xb4S\xb8\xf3\x98\xe7\xaeB\x93\xf9\x89\x1a\x8e\x896x\xdbi\xa3[\xe1\xa2\xaefq\x91\x8f\xde\x8f\xd0\xde\xfa\xa6\xa0\x83\x8e\xb6\x92AKZr\x81 \xed~\x83`\xb1\xf4\xd0\x96\xfe\xacv\xb5H\xc9\x8fx@\x8a\xfbl<\x81\x92\xe4\xd25\x8e[\x0b\xd6\xe6\x84\xaas\xf3|4l\xa2&\xd1u\xb2\x16\xfb\xaa>\x0c$\xef\x03\xa0s\xa2\x9am\x1d\x06\xa5S\xad\xaf_\xfd\xd1~[\xc2\xda;\tW\xdb\xc2\xc0\x14\xfe\xbf\xf6i\x04\t\xde\x06l\xea\xde\x9fG\x8dE\x9fe8\x80j,:4\xe9H\xb6\xd7\xaa\xa2\x94(\xd0\xb4\x17\x9f\n\t\x8f\x81v\x93\x809\xab\xb2\xeexHL.\xba\xe8\xa5\xc3\x97\xa3I\xad.\xf0\xd4N \xd2\x17\x8c\xfc\xd4\x12a\x93\xfe\xc1,\x15\xceb\xcd\t\x86y\x8e\x12\x1f\xc6o\xe4\xe5wx\x84\xadmZt\xd5\xd4\xd2\xb7\xa8\xac\x0e]\xfd\xf8\xa3\xe4\x86\x82\xc7d\x1e\xc2\n\xcb\xce\xf7V\xc28k\xd1\xefV\xd4J\x7f\xc9\x14\x90^E\xef>\xdbA\x10\x01vS\x90H#\x90n\xc1\x85\x18\xfa\xa2v\xd3\xd9\xde\xf9U\x89&\x03\x8b\xcf\x90\x8f<G\x80\xfd\xb2\x18\x18o\xea\x91G\xda\x98\xd9\x89:cq\x03h`C\xb4\xecHc\x98\xf5C\xf9\x84\xfbz\'5\xd9\x88\xf7\x8c\x15\xe6\xe9\x10\xc9\x86,\xd1WRN\xb8MYm\xf2\x175\xddL\x03w\x8fE\x89\xd8=\xb2b\x02\xd8]\nf \xd5\xf6\xac6\xdc}wp\xe6\x19\x8c\x90\xb1\x1d\x9b8\xb7WcX5\xa3D\xe3\xe4\xfdR;O\x08*\xc2\xcf\x93\xb3\xe0Gq\xc9~~\x91Q~Y\thK\x04\xac\x07v\xe5A6|\x04Gr\xde[\x04\x13\x8a\x88W\xfe/\xe8\xefL;=\xc8!\xb1\xb3\x03A\x9a\xba\xa1\xb7\xff>\xbf+u\xaaK*/f}Y\xe3\xb5\xb2\x81\x01,>Tx\xd3\x06\xfc\xdd\xe3\x1al',
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="google_unknown_tool_id",
                            name="get_book_info",
                            value="Title: Mistborn: The Final Empire, Author: Brandon Sanderson, Pages: 544, Published: 2006-07-25",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"title": "Mistborn: The Final Empire", "publication_year": 2006, "author": "Brandon Sanderson", "pages": 544}'
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "title": "Mistborn: The Final Empire",
                                        "publication_year": 2006,
                                        "author": "Brandon Sanderson",
                                        "pages": 544,
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b"\n\xb3\x08\x01\xd1\xed\x8ao~\xea\x01J\xc0\t\x15Z\xc3\xf0\xf8s1BCsl\\\xd2\x99\x85\xfc\x12\njW\xb5Z\xda\xcaq\x8d\xd7c\xfe \x0e\xffIn\xa0\xed\xc6IN\x99Z\xe44<2\xb8\t\xfa\xado\x83H\x81\x86\xe0m\x07h\x8f~&I\xb5\xd1G\xa2\xa7\xbb\xb1?\xed\xbd\xe0\xaa\xb0\xe1\xb25\x9b\xde\x1a\x19G\xb4\xd6s\x89\x06;\x98\xf7\xe1\xd7\xae\xb8~as\xb7\xcd4G\x9f\xd7\x04\xe4\x19\xceCV&\xe6+\x90\xcd\x8b\xfc\x12+\x12\xb3\x1d\xda\xecw\x99\xcbp\ng\x83M\xc4\xbeu\x97\xaa\xd5\x1fo\xda\xb8\xff\xaax*w\xd8\xf4$j\x80Q\x1d\xc9\r\x01\xe6\xfa>)\xe0\x12\t\\\xa9\xafS\xfc\x91\x05\x04?\xb0S\xd0\xcb6X\x88\xcf\x81\x18\xed\xaa\n'\x16\xe9\x7fG\x97\x92\xb9Y\x06n\x08\xea\x90\xf0i\x0e\x94\x8e\xdb\xfd\x0e\x9d\xbf\xa3\xc0\xe9\xbd4#\x86\x8a~\x07\xfd0\x80V[\xd7M7\x02\xae\x85\xe3\\pJ\x97\xdd\xc6Z\x17}3/\xf3ZdG\xc8\xeb\xe1\x91S-\xf7\xbd)\xb7b\x1b;)\x0e\x993P\xc4E\tG\xc2\x88\xc3\xc0\x9e\x16\xdc\x0f\x1c\x17P\xd8\xf6\x9c\xc8\x90A\xae\x8d\xcd3\x8d\xfc<x\xa7\xcbo\xafk3\x8af\x1c)\xea\x97\xf01ZD\xfer\xa2;\xb5G\x13\x8ek\xdft`\xd0\xf3\xf9<2\x7f\xf9&\x9a\xcc\xaa\x06{\x12\xac\x8fq\x15\x1cD\xe3cNQ\xbeoa\\\xe3\xd5\xe1\x1b'\x8f\xea\xe7|\x19\x9b,\xd6\xd0&\t1}\xee\x15\xa7\xc5L\x0f\xe8B\x81XD\x97\x1f2\xcf^\ty\tUe\xab\x94\xac\x05;\x91\xc8n\xfd\x83$\xaf>R\xf9N{l\x183\xa6\xf4\x11\xa2\x87La\xfd\xa2JK\xb7Z\xa9\xa1\xd6\xa0\xac@b\x7f\xcf\x12\x1fUL-\x1aDF-t\x1a\x8e\xa46Q\xc5.KH\xb3f\t\xe8\x00i\x80o5\xd1\x92\x13\xa52\xbf\xb6\xa0\x1d\x0c}j\xa7\xe7\xa2\xf0\xc9S\xb9\xe8\x15\xfbH\xed\xdb\xe9\n\xb7\xf0\xb7\x84\x81< lZ\x80vW]\xfc\xc9l\xdd\xfa\x91r\xd75i\xc4d\x19\x81\xd7\xe3\xd2\x14'\xe3<\xe39\xdf;c\x19\x15\x05\xa5\x9a\x187\xb4\xa6\xd7\x0e\xfb\n/\x1e\xff\xea\xc3\x1cg\xcak\xdcHf\x1a\xcb`\xcf\xc2x\xf2@\xd9\xba\xd6T\xb7Hwl\x01\x83R\x13\xfd\xd7\xa5\x83#>\xe3\x80t\x86\x81\x96\xcb!\xadX\x06/\xd7\x96q\xdc#t\x0b\x0c\x1bMg\x95n\x1a\xdd\xb3\x8c\x10\xa1u\x90\xb8t\xb9;G\xade\xc6{j\xd7\xf0\x84\xa1^\x8a\\d\x97\x10k\x95\xe8\xe6\x1c\x03\xf1\t\x14V-\x84\xd0\xac&\xed\xdf\xba\xe7?p\x1b\\\x9eJj\xe6\xce-\x8a\xa2\xac\xae\x01f$\xd3\"\xfaH\xccD\xe7\xe8\xf3\xcc\xe7\xc1'\xba\xbd7o\xe3\x95\xc05\xa5\x91\x8d\x96\x0b\x0c\xed\xab\x1e\xe0v\xadV.M\xbf\\]G\x0fu\xf1u\x976Qau5K\x10i\xbf+V!\x05o\xc4\x9f\xe0\x9a\x15\x06Qx\xfc\t\x8d\xb2\x95f\xe5\xdb\xbf\xe8FXhBkT\xd6\xfe\xf2\x8e\x8c\xb6g&~\xc6\xdb\xfd\x82gkl\x14\"=\xa8Hq\xa0\xadaG\x12\xc5\xb8\xd1Hz\x8b\x83\xc4\xb8\xa9c\x01!x\x1e#\x15z)\x1a\xdc\xa9'\x1aB\xcb-\xac\xdc4\xd7\xdd6\xf9\x14\xdb\xe0^\xffX\x9b\xa6\xe1F\x98\xcb\xa1\xc6l+k\x87\xade&\xa8\x90\x8a\xcd\xcel\x118\xc9\xc9\xce\x98j4\x1eK\x1f\x10\xf9M[m\xb7\x81\xe7-\xab\x00C\xee\x16\xf1gfK\xb3;2\xe6.,\x81\xa3\nz\x89\x06a}5\x1a\x96\xf2i\xb5l\x91\xac{\xed\xf5m\xb3\xb5S\t\xcf\xdd\x99\xf5/\xec\x8b\xd0b\xed\xa6\xa5\xffc\xd5p.\x863^$M\xbd\xb1C\xcb~V\x93M\x1f\xae\n\xec\x02n\xe1\xe1\xdc%\nT\xf8)\xdc\x13|d%\x82\x80\xaa\xfae\xb3^A\xc9\xb9\x88o\x96\xa0\xea>\x84\xc4BZ\xbb\xb3\xcd\xf2\x98\xe1fs\x18qs\x18\xc1\x8e\xc7\xc1M\n\xd0\x02V\xd3\x85I\xb7\x84v\xc5\xc0\x04$\x86\xad]\xfe\xd5\x8c\xfd\xf8\xfd\xdcS\xf7\xf7\xf2\xbcMJ\x87\xf8\xe4\xb0\x17I\xe7\x9e\xe81pZ\xd4RW\x13\x95\xfaH(\xf3w9\xc8P:\x8dhTB\xf8]\xcd\x83}\rM\x89\x8dK\xcbR\xf6\x8e\x98DD\xfb\xf3O\xd9",
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "BookSummary",
                "description": None,
                "schema": {
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"title": "Author", "type": "string"},
                        "pages": {"title": "Pages", "type": "integer"},
                        "publication_year": {
                            "title": "Publication Year",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "pages", "publication_year"],
                    "title": "BookSummary",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [
                {
                    "name": "get_book_info",
                    "description": "Look up book information by ISBN.",
                    "parameters": """\
{
  "properties": {
    "isbn": {
      "title": "Isbn",
      "type": "string"
    }
  },
  "required": [
    "isbn"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider": "google",
            "model_id": "gemini-2.5-flash",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please look up the book with ISBN 0-7653-1178-X and provide detailed info and a recommendation score"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="google_unknown_tool_id",
                            name="get_book_info",
                            args='{"isbn": "0-7653-1178-X"}',
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {"isbn": "0-7653-1178-X"},
                                    "name": "get_book_info",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b'\n$\x01\xd1\xed\x8aod\x85\x91\x16\xff\x0e\x96\x94\xae\x18|\x1eV\xf5\xb8\xe3\x88\xd5\x86\x8d1\xb5D\xc3_?\xe9\x83\xc3! \ny\x01\xd1\xed\x8ao\x85\x86.C\xad\xfb(\xe9\xd6C\x8e\xfc\xcdn%\x9f/n\x02\x0b\x01w\x05S\x81e\xdd\xcd\x99\x16i\xfd\x13w1\x11\x8e\xb0n \xb9o,\'n\xcb\xd3  \x99;P%*\xda\xee9\xc2\x12\xe0\x88\x81.\xff\xc5d\x91P}d}\xff/\xcc\x08\xeeI^\x14\x8d\xa3\x01\x9a\xc5\xe3\xd1s 2\xde\xd9\xfd\xc8\x9f\xc36\x96\xfc\xa3\x8dHu3\xb5\xc5\xfc\xcb\xc6\x88\x90\xea\x80$\xb1\xab\x13\n\x82\x02\x01\xd1\xed\x8aob\xd4\xaf\xa8U+\x1a\x0eW5)s\xc2\x1ck\x8c!\xe01\xe8\x83\xd2\x19\x86\x970~.\x8a\xfb\x8d\x8a\xd7\xef\xb0\xd7\xb0\xc5^[\x1bU%Jf\xba\x1bv\xef\xe7\xec\nT\xa7\xb2\xf5\x0c\x8c\x1dQ\x89\xee^>\xaf\xb3E\x13\x8a\r\xae\xe12sE\xe0\x8et\x18\x81\x89\xcec\xd5zC\xe9\x171H\xdc\xc7\xac\x1cH$\x1a+\xd8`8g\xbe\xf4\xe2\x0c\x13\x90\xd3\xac\x01tU\xfc2\xdfH\xfa\x9e\x92!n\xa8\xc79*\xc8V\t\xb0\xb6\x8b\x7f@\x12\xe7\x9f\x8a"w\xab3\x03\xb6\x83\xa6|\xb4s\xacY\xd57\x17\x83\xd1T\x1f\xe3\xbc\xbbq@\x0c\x8fw\x81\xe8)/\xc1n8yk\xce\xe3\xaeg@\x99\xaa\xb8\x12\xa5K\x04q\xf7H\xcc\x1d\xd7\xec\xa8\xf9\x0b\x1c\xc5\x8b\xbc\x83\xe9Q\xed\x10G%\x9bSWXe\xc75V\xcb\xe4\xe4t\xb2L\x02\x1c\xd2\xf9\xb0\xa9#X\x0f\xefw>\xca\xc9\xd4\xac-\xe6\xbe4i\xacw\xb7\x0f\xbcn\xa3\x19\xa6\x08\n\xc3\x01\x01\xd1\xed\x8ao=\x06\x1az\x15T\xadH"9\x94G>\xe1]\xe6\xeeD\xe5J\xe56\x8a\xc4\xf2\x9aE\xd1\x18\xe3E\xb5\xac\x96\x9c\xd5\xde\xaf\x0e!\xb3\xa4\xca\x06\xdc\x18\xd0\x8d\xd8Q\x14v\xed\xa2E7\xba\xc9B\x1d\xa7\xeb\xd7\xe8\xce\xef)Ar\xb2\xb2\xa0C^\xc2t\xe9\xfa7\xa6\xf4?V\x97\x17\xdb=\xb8\xafh\r\xa1y\x9e\xd7\xf5b|0L\xa80\x13\xc7\xa9/\x96\xff\xe5\xe2\x0b\x11d\xbc\x05\x9b\x9b\xae&f\xe4b\xd3\x12\x9f\x8e\xe4\xbb1\xaa\xa5\x18-\xeft\xab\xa3\xf7H\xaf\x95\xedC\x03\x1b+m\xac\x08-\\\x90\x00\x10\xb9\xb2\xe8\x12\x94\x00*\xb2\xb3\x80\xe9\xde\n\x80h\x85z\xc3\xa5G\xf36^m\xd4\xf0\x17\xd8\xc7\x9e\x82\xf91\xde\xd6\x94',
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="google_unknown_tool_id",
                            name="get_book_info",
                            value="Title: Mistborn: The Final Empire, Author: Brandon Sanderson, Pages: 544, Published: 2006-07-25",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"publication_year": 2006, "pages": 544, "author": "Brandon Sanderson", "title": "Mistborn: The Final Empire"}'
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "publication_year": 2006,
                                        "pages": 544,
                                        "author": "Brandon Sanderson",
                                        "title": "Mistborn: The Final Empire",
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b"\n$\x01\xd1\xed\x8aoI\x13/\n\x9e\xaev?\x1e=\x81\x04V\x93\xe2C\xcb$Bm\x1f\x94\x8e\xefGQ\x9c\xe1\xb1\xf1\xee\n~\x01\xd1\xed\x8ao\xe6P\xfa9\xc5\x1f0{\xa6\xdf\x16\xe2\xf8*kh\x9b\xb7=\x91t\x84Z\xf6\x93\xdc\xd8\xa7\x1e\xc7\x87c,^\xf0{\xeb\xf3\xab\xc9\x8f\xa0\x0e\xe3 `I\xc0\x93\x92\xcb\xbc\xb1t\x9b\x8d%<I\xe2\x01\xa0\xc8\x97#iZ\xb9\x87K)\x04\xc1u\xfe\xe5\xf6?6\x81\xea\xb1TCx\xeb\x11\x19Y<L6\x08\x0f\xdb\x9f\x9d(+\n\x1e\x96\xc7.\x96\x11\x10'\xd0\x1d\xce\xbf\xd9x\xb8\x1f\xe8%\xf8q\xdf\n\xbb\x01\x01\xd1\xed\x8ao\xcdE\xe5\x90\xcc\xb7\xf5]\xd9=:\xc9\xa2\x16\x8a\xd6}v\xd4\xdd{\x88\xb6\x1f\xa7\x82\x0eN\x16\"\x07\xec\xf6Y\x01\x1a\xc8\x9a\x0f\xd4Zv\xfd\x06J\xb4\x99\xcae\xe0U\xf6\xb5h\x02\x17\x84\xbd\xc7e\x9c\xa3\x10q\xaa&\x8b\xc5C\xef\xc5\xf4\xa39/\xb1\xf0\xf3\xdaa\xe6\xb1\x88[\xb4d+\x1cZ\xb4\xa4\xed\xa1\x0bW\x93\x00K%\xdf\x99\xb1\t\xa3c\xe8\xf8u\xffGMmV\xb5\x9c\xd2\xd8\x06\x7f\xaa\xebm\xda\xc7}\xd4\xf0hHH4\x8f\xd5\xc2*\xe1\xddy\x86\x1aM\x19\xf4]\xfe\xdb\xe1\x8aU\x8c%3_\x12\xb0T\xc0O\x9c\xa8/v2\xa7<\xfcP\x8d\xa1\xa3\xd4\x1bi\xcb=3\xde3X\x1a\xd5p\n\xb9\x01\x01\xd1\xed\x8aoj\x9d\xc2\x13\xb8\xad\xd9\xb2\xe0\x00<\xf3\xefVi\x02\xb91\x9b\xf0\xd2V\x86\x0fz*\xe7\xb6\x94\xe6ZhbP0\xeb\x90Z\x0e\xb8/T\xe68\x96q\xe7\x83&\x83L\xe7\xe5S\x86V\x82 \xe0\xc2\x9a'\x9bNI\x16SG\xa5\xc8W\xa5\xef\xa7\xec\x1d\xec\xf8Z+\x020)\x11\x97\x013}\xd3\x08H~\xc3[\x14\x9e\x97B!g[\xb1\xabN\x8c\x11\x81\x16\x92\ts\x89L\xc2s8\xe1\xa37O_<s\xff\xb8-oLG\x16\\\x88\xadA\x06\xdd\x15@h\x0e\x87\xae\xb3\xa1^\x97\xeb1\xdb\x80\xcc7|\xda\x00:t\xa5\x89\x07v\xec\xd55\x842\x82o\x14gr\xbdj\xaf\x88\x11\xfe\xe8!\x85\n\xca\x01\x01\xd1\xed\x8aoo)\x07\xcf\xecP\x06\xa5\xcaJqL\xac\xba\xa0\xb0\xb5\xac_\x89F\xb3\xf5q\xb1\x8c\x06~\x01\x7f\xd8\x08\xee4\x9cD\x89\x92U\xc7\x048\xeepp\xca}\xeek\xbf=\x10\x1f\xde\xc6\x88\xb5zy\xb9\xe5\x91\xc4\rq\xd0]\xc0\xaa\xd0\xa2/\x94\xf8)\xee((\xc6_1\x13\xde\x11\x82\x0f%\xb3\x98\xb9N$9\xb23:\x87c\x93\x15\xad4-B<\r?='\xf9\xb5\x1dWF\x85\xa7\x99\xb5\xa9\xc4\xed\xf2np\"\n\xcf\x9d\x9e\xfd\xd0\xa6\x8ey\xe7\xf0\x17at\x91\x11@t\x8b\x17Xnr^\x9fa\xed\x8d8\xe7\xad\xb1) \xa2\xd8\x82\x0b\x9cJ\xe0r\x90\xbe\xfde[\x81\xe6\xb8\xee\xb2\xd3\xbf)\xa8\x9f\xfe1\xf4^\xd9\xe9(\xb1\x01\xad\x05\xbc\xfe{}\n\xa5\x01\x01\xd1\xed\x8ao\xea\x1b\xd4\x0fc\xf5\x96%\x88\xbe\x96\x83R\x9d\xab\xb3\xae\xfc\x05/bZ/\xbcu\x80\xe6]De\x06\xbe\xb2\xc9\xa2\xf5\x91KG\xe3Q\x03&\x82c\xc8\x9e\xc3\xf76\xc8\xea}\x16\x8c\x05\xfay\xd4X\xa4SGOL\x8at\xe4\xc9g2\xf1\xcc\xf6}\xe50\xb8\x1b\xf6\xeaY\xda\xd4\xd4Q\xe2\x97\x8f\x08\x90.i\xdbG\xaa\xe4\xd4\xb4Y^\xdb\xf5\xa5\x8eYDo-&a\xd8\xc0\xff\xc2\xc9?\xaf\xd7\\\xd9C`\xce\xfa\x1a\xbc\xf7\xd6\x15UIwP\xca\xa6\xe2#\x07\x8d#\x07.\xe8\xa9E\x80&\xec\xe8\xb3\x19\xa4\x80\xc8\x93d\x12\x93\xd8\n}\x01\xd1\xed\x8aoq\x1a\x06G\x8d\xf4qfay\xf6\xc7\xd3\xf3\xea\xaaeJ\xc3\xbc\xc5g\x9c\xb1\x8a\x95\x85TF\xbcB#\xbf\xdasp\t\xc0\xfb\xe0>\xc8\xd8\xfe\xfcP\xe4QO\xfa\xd4\x13\x9a\xd6\x0b\"\xc9G\xc28`\x89t\xd7\x19\x01\xd2\xd3\xea\x89n\x1a6#\xd96\xe5K\xb9_%\x1d\xb8\xfb?C\x94\xb3\x96\x99:\xac\xb9\xad\x05Z\rL<\x92\x9a\xf1\ri\x7f\xa0\x98\xd8\x02\x12!\x97\xd3\xebN\x0f\xae\xfbf\xe1",
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "BookSummary",
                "description": None,
                "schema": {
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"title": "Author", "type": "string"},
                        "pages": {"title": "Pages", "type": "integer"},
                        "publication_year": {
                            "title": "Publication Year",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "pages", "publication_year"],
                    "title": "BookSummary",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [
                {
                    "name": "get_book_info",
                    "description": "Look up book information by ISBN.",
                    "parameters": """\
{
  "properties": {
    "isbn": {
      "title": "Isbn",
      "type": "string"
    }
  },
  "required": [
    "isbn"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
            "n_chunks": 3,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider": "google",
            "model_id": "gemini-2.5-flash",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please look up the book with ISBN 0-7653-1178-X and provide detailed info and a recommendation score"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="google_unknown_tool_id",
                            name="get_book_info",
                            args='{"isbn": "0-7653-1178-X"}',
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {"isbn": "0-7653-1178-X"},
                                    "name": "get_book_info",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b"\n$\x01\xd1\xed\x8ao\xe1)\xeb\x02n\xad7\xc2\x8bQ2\xdd[\x1f\x06\xa8\x039dUa\x0f\xb9\xd0*\x81ng\x0e\r\x8d\n\x80\x01\x01\xd1\xed\x8ao\xa7\xac\xde\xd1#\xbbL\xe8\xf2\x1fs\x85>\x9b\x04\xe7\x7f\xb9\x90p\x15\xd7\x9e\x01{%s\xe3G\x191w\x12\xd1\x9a\xaa\x9d#\nz\xb5\x19l\x81\x82\xb6\xfb\x84\x88\xf1B?#7\xca\x89\xd6\xd6\xb4\xbfZ\xf2+\xb2\x1a\xf6\xdb\x8dj\xf4=\x89}m\xe0\x9aHRSo\x06\x04K\xbd\x1b\x12\xed\x80\x8568\xa4\x7f`\x80\xa1\x8b\xe3&\xaf\xd0\xa5E\x9d\xc0!ei\x83Z\x00f\x18a\xf8)\x0c!R\x93m\xf7v\n\x96\x02\x01\xd1\xed\x8ao\xba\"\xad\xef\xb4\x8b'\x9e\\(\xa7\xfe.\nh\x95s<\r\xa5\x0e\xedxp`oF<\xdf\xc9JrQBs\xee\x81\xd6\xc4P~\x05\xc6\x8fO\xc1.\xca\xff\x11\xad\x1e\x95)\xa8\x93[\x8e\xe1\x05\x80\x16\x1c\x86K\xf1\xee\\\x93\xf7(\x80\x03(j%2\xf9\x05\"!\x96\xd2\xdb(\x82@\\8\xcd\xff\x1b\x8ac\x06\xbe\xe6\x8b\\Rc\xd4Tn\xd9C\x13\xe5i0v\x92\xf0\x7f\x0c\xe3\x84\x1e\"1\xca\x87\xd7_2-\x1e\xd2\xeb\xc1\xfbq\xcc\xf7b\xe0\xfe\xeb\x11+\x01\xe0a\xa8.\xb1\xfd\xea\x9c>\xd0\x01\xb0\x19G\xf1\xb1\xdfl\xe7N\xbc\xf1\xff`\xe50\x07\xdaa\x93c\x17\xdc+/$\xb6\xf9^\xe0Vv6\xd4R\x8dg\xc6\xf5\x91r\xab\t\x8d\x0e\xebM\xbb\r\xbb\xeb\xffd\xd73\xc1\x00u\x9b0\xd6\xff\t\xc1j\x979m\x18\xb8oNND\x99\x91r \xfd\xc2B\xf6\x87\xbcu\x9b\r7\xb6C\xb8\x19`;\x01\xb0\xcd\xa3\xda\x0e\xd1$\x9cu\xc24\x18@1\xa8\xf7L\"]\x05\x83S\xcf\xb7\xe41>\n\xbe\x01\x01\xd1\xed\x8ao\xd8[k\xe0B1n\x0f\xc5\xaaB\x17\xd2\xe3E\x10\xe3bw\xdb\x07\x92Uy>\xe70\xba\x1eh\xc1\xff#Uca\xe8\x7f\xbd\xd9h\xf9BV%\xbb\xd4\xb3\xfd\xbfFN\x97Q\xe2;\xa9*\xd6\xae\x9b\x0b\xe4\x15\xcf\xe9_\xb2\n`)\x90\xa2i\xf2\x10\xdc\x05F;\xe2\n\xea\x18\xf1\x81\x8d\x92\xdb\x93\xe6\xee\xb3\x84\x8a\x05\x0bG\xaa\xdfeM\t\xeen\x8a\x0c\xc6EyDHQ\xc9\xdf\x9c\xbe>\xb1#|\x83u%\x86\x9dg\xaeX\xd0?\x0c\xf5\xea:\xc3\x81Bof\x1eOR\xfd\\\xbe\xa1\xe4Q \xee~\xcb\xd6m\xc3,\x9a\xc2Q\xfd\nwy\x81/\xd1\x85\xdcW|\xd6x\xdfcm\t\x94Y\x17\xf2\x81\xf4\xd9m\n\xdb\x01\x01\xd1\xed\x8ao\x16\xe9\xbf\x9a\x14\xfb\xa0\x14\r/i_`W}\xdd\x91D\x90\xfb8i\t\x8e\xbe\xe7bg\x8e\x12\xeb\x084\xab\xd9\x95\xab\xfdSN\xc7\xc7\xf1\xf9\x16\xde\x8d\xea\xbc\xe6\x93;\x94\x8dv\x00\xf8\x91\x85t\xaet\xc2\xaf\xa54Lv\xdcA\xf6jm\xc5\x8f\xc8\xc2 iUk!\x83R\xce\xfc\x99\xdc&\x93H\x02\x00 ,\x805\x05\xcb\xda\xf7,\xf9\xc1\x0f\xac'@\xb6FNG.\xf7\x94\xd8\x17\x80[}\xc6\xb9)f\xca'\xa6\x1bx\x961\x98:\x11\xa2\x91ToE\xe3d\xdb{g7\x90\x08|\xa8\xa3\x9fm\xfb\xfbU1H\xb8\xbdr\x9ad\xe9Nk\x95\x10\xc2iO\x80\x89\xedD\xb4ih\xf7\xbe\x94\xa3\xf7yE\xdb\xac\x0e\xcb\xea$\x10d\xeb\xf3\xaa\x1f\xae\xa3\x18\xa9t\xc3\xfa\xc9?k\xcd\xf1\xe7\xc6\xce\xa1\x89F\n\xd1\x01\x01\xd1\xed\x8ao\xb8U8U\x85E5O\xbdz(\x13.\xd3\x83\x0e\xce\xa5\xc6'\xc0\xf1Ro/\xac\xe6Q\xf3\xd3\t\x11O\xb4\xfeK\xab#)\x9d\xeb\x03\xc6\x91\xe7\xcfI\x84\x975:\x8e\x9e\xd7\x08\xb6\xa6\xf0\x00'\x86Pz\xd3\xa2\x97n9NJ4\x9d\x98\x8a>\x9bM\xb7\x92pR\xe9\xe8\xf0\xd3\xaf\x00\x96\xdd\xcc0k\xb7\xbe\x15\x02-Wu\xfe4\xa1\x91\xce\xe0`x\xca\xcf#\xfa\xee\xdb {(Q\xf6\x87\xbfp\xe1j]00\xae\xb5\x88+n\xa3\xbf#\xa4\xd9\x00\xfc?\xf4k\xe3\x81\x1a\xcf\xc5i\xa2\xaco\xf7O\x15}\x81\xfe\xc2\x91U\xb0\x9e\x7f\xfb\xfe\x9d<p8\xd4$_\xcf/\xfe\xc4\xa0\xea\x8c?\xbc\x10\xe46}\x02\n\xc6\xa9<\x9e\xa7\x83\xa8\xde\xdf\x16\xba\x0e\x90'\xa7\xaf\xde\xa4\n\xa7\x01\x01\xd1\xed\x8ao\xeb\x1b\xe6\xd2U\xcf\xf6\x15y\x85_}\x90%F/\xb6\xd1\xb6\x8c\xc6\xd2x\xaf[8\x1d\xca\xe1\xc3\xf2\x01B\xe5[\xd7\xaf>\x04;\xb24\x15\x06\xb7\x83\x9b\x06\xac\xfcH\x9d\xbd\xed\x1b\x06\x00\xa4\xf3\xa9\xfd\x1f#\xc1Tj\xe0#\xd5\\\xfc\t=\xbf\xf9hp\x11\xa1\t8\x84\x94\x0b\xbe\x05H\x13\x8d\xce|g9L \x91\xce\xddu\x9f\x9aO\xed{+\xb9\x13_'\xc1N\x9fB$fL\xe4N\xf78\r\xa2\x84\x88\x89\x82\xfe\x03f\x1c\xbb?\xeb\xdf\xaf\xeb\xd4\n\xd1\x94\xf9\xa6\xc9\xe3c\xd5f\x90\xbb\x8a\xa9\xfeK!\xd6\x9c\x10od\x93\x90\x05\n\xee\x01\x01\xd1\xed\x8ao\x01\xb4)\x00\xbbX\xbb\xd0\x8f\x8f\x17\xb9\xfaS \xd1@^\xd3\xc5+\xfb\xd5\xf0K \x93\nNe\xad\x12\xec\x11j\x98\x0eF\xf9\xfd\xeeUlj\x0f\x02\x81:\x01\xd4\x87\"O]q\xf0\xf5\xde\xcd\xf2I/\xa8:\xe4f\xa9y\x15\xe2:\xf7]3@\xfan(\xec\x0e9\xc5\xe1T\x08\x1e3\x83+\xb4)\xff\x93\xb8\x98\xca;\x99\x94\xde\x80\xbb\x024\xcf\xa4\xb2\x85\x83\x8e\xb6$\x81\xa4W\xe6\x17W\x94\xb7\x01\x12\x90\xe7\xe6K\x99\x90\x9c\x0f\x06t\x1f\xae\xef\xf7{}\xd9\xfc3\xc2\x13\xef\xbb\xbc!Vr\xeei\x91\x16\xcaO\xda*\xdc\xf0\x1fL\xd2'\xbf\xd6'\xca\x15\r\xc0\xf3\xebQ\x1f\xc2\x9e\xfe\x8b~4V \x99\xcc\xe7\x99-3\xf4\xfbT\xad8u\x06\xf8\xb5?\x05\xc3\x03\xe4\x043\xf1\t\x8d\xa9 ,\x02\xef\xb2\xc0\x85\r\x0c\xdd\x1c\xf3\x91\xd8\xce\x15\xa1\xa7L\xd5\xa7yrL\x13\nj\x01\xd1\xed\x8ao\x0c\xa776@\xefw<\xde\x87\xb8\x048A\x08\x92\xce\x05\x15XIY\x8c\xdf\xfc\x92#\xb5\x1bK\x11Z\xdb\x8d\xf8dC\x01\x03\xb6\x9c\x1b\x89\xfc\x80\xa7\xe0\x98\xabW/\x840\xd1\x06\xdb]\x00\xdcM\x01\x81\xe4\xce\x9fBp\x1aM\xc3\x15\x86oM0T\x06\x86u\x85 \x9d\xfbR\x98\xd4\x7f\xc7\x92B~7\xb0\xb4e\xc2\xf0W&\x97\xce",
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="google_unknown_tool_id",
                            name="get_book_info",
                            value="Title: Mistborn: The Final Empire, Author: Brandon Sanderson, Pages: 544, Published: 2006-07-25",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"publication_year": 2006, "author": "Brandon Sanderson", "pages": 544, "title": "Mistborn: The Final Empire"}'
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "publication_year": 2006,
                                        "author": "Brandon Sanderson",
                                        "pages": 544,
                                        "title": "Mistborn: The Final Empire",
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b"\n$\x01\xd1\xed\x8ao\xe6!p\xf7 i\xe7\xd5\x8b\xa0\x95\xfc+[\x1d\xb1\x90\xca\xce_\xf8\xff\xae\x8e{\xfc\x10e\xc0\x9cr\n\x7f\x01\xd1\xed\x8ao\xce\xbfi(\xdcs\x9e\x10\x0c\xa2\x10\x9ez\xc2\xf7<i\x81\x0ci\xcd\xc0\xae\xe7_\x90\xce\x04\x80\xa6o\xac\xbf#\xd0#\xbb\xc1p\xc9\x11g\xfb_A\x15\xf9T\xe6\xf4\x8c\xb9A\x0e\x10\xd9\x8d\xcb{\xb6V\x06$\xf3\xd9\xdfFQb\x86\x14\x8f|Awq\xe4\xc2(\x9b\xe2\xbd\xc6M\xd2y\xfc]\x08\x97n\x0b\xad\x12\x9fO\xb8\xd6O[\xf3\xadi'\xe5\xcak:5\x9a\x00\xdc^\xf7\xab\x05\xf8\x94\xb0\xf8@\x18\n\xb9\x01\x01\xd1\xed\x8ao\x1d\xd8\x1e\xe8VK\x18\xb8\xbd\x96\xb1\xe6\xfc\xf6bI\xf4\xdba\xfeZ\xe5\x1c\xd2\x12P-\xe8l;tv6 R^\xee\xbd`\xf4\xf2\xea\xdd\xa4\xfa.\x190\xc1\xbf\xaf\xf2\xe9)Q\x80\xd99\x96\xf8\x91\xfb'\xffQ4\xa2[\x03\x84\xd8R\xdc\xb8\x12\x973K\t.;\xc2\xdf\x9c\x8f\x8a\xa1\xb0P.KI\xa70\x82Q\xf6\xaf\xfb\xa9}\x8b:\x8b~nn\xf0\x0f\xf0\x93\xf7+\xde\x94pL-\xf6\x17\x0e\xd4\x05\xedC\x02\xb4\xe6\t\xd5\x1f\xba2\x00\x8cj\xcb\xcdq\x18\xe0~\x06\xb1\x84\xbf]M8\xcb\x05\xdad7\xf7\x86\x0f\x8e\xd00aO@\xcd\x90\x8a\x9d|\x8b\x91\xd6\xf5&\xcf8\x80\xfb \xbc\x95\n\xac\x01\x01\xd1\xed\x8ao\xdbg\x8e\xd9t\xa3\xf1\xc4>@$u\x8f\x88\x95\x11\x8cX`\xd1,]\xdd\x84B=\x87\x90\xcbX\x1c\x1cJ\x7fA\x9e\x86Z\xdc\x84\x13\xd5\xa2\xfc\x8b8\x8em\xbdP\xf9r\x82+\xcb\xaa\x15\xa4.\xd0\xe9@J\xec\xc9\xc9k`\x9e\xbfu+&\xe7\x96\xc9=B\xce\xdf&\xa4\x9f\xe4\xef\xd5\xc9\xf8\x0c_\xb0\xd5\xe2th \xbb^\r>$l\xc4P\xeb;\x85\x0fJJakH\xce!\x19W\xd4\xb9\xdfU\xe6\x95\xf8\xa8\xc2 Z\xd2Fy\x87\x98\x9c\x1e\x04C\xd1l\x03G=^\xb6\x94R\xaaG\x99\x029\x00`\xede\xbd3x\xfbk\x0f\xb8\xc1\xaf\xe6j\x1b\n\xf0\x01\x01\xd1\xed\x8ao{\xf7oI\xcc\xedLx\xf6\xa1\xbe\xe7B\xe5\xc1\xa6\xc9\x8bq \x80\x1c\xef\x08\xf6\x8f\xbb\xc2z1\xc4U\x08d\xfa\xf3\xf4U\x9c\xd6\xf21T\xdd\xd1\x00\x9aBT\xc7\xfa\xc3\xcc\x12\xd3Bu\xbd\x94n\xe2>o\xff\xd5\x9b\x82\xe1M\xbf\x1aS\x03\xee\x087M\xf2/j\xe9\xf3f\x0b\xd8W\xf5\xfda)^b~u\x9fc\x956\xc2\t\x05-\n\x9b{\xf6\n\xc8\x95\x94WO\xb3\x0e\xa5\tI\x91\xa51!\xa8\x9ch\x18\xc1H\xa6\x83\x8d\xad\xf2\xe3\xc7=I\xa5\x8fnk\x0c\x8d+\x16\xe2\xf3\x1633ane\xa7z\x9fA\xbf\x98\xfaW\x90\x1c\x83\x97\xe9\x97A\x9d\xe9\x9f\x9d\xc8V\x11\xebNA\x01\xd5\xdd\xdc\xb1K\x88v\x11\xe7\xa6\xd9\x89\xd8\xa7\xf9\x9fo\x9bwb\xe5\x0b\x99\x15\xec\x05F\xc8\\\xebaS]\x80\x84\xf2h\xcf+LU~\x13\xc1\xf0\xae\xe5\x14\xf5\x90\xec\xe8\xeb[\xbb\xfe\xf1\n\xaa\x01\x01\xd1\xed\x8aon\xf2\x81m\xcbk\xbc\x8e\xd1\xf2\xf8\xf4O\x18j=\xc8\xf7\x83\xe4\xef\x07\xd2\x07\xdfg\xd8<Xu-\x12\xab\xaat9\x8a\xfb\xd2\xc4w\xc1\xd6\x02I\xa3\xcb\xc2WVN`.6\x95\x10\x9ai\xf4\xaf\xdb\x17B0\xfd\xf2R\"\xee%\x0e\xdep\xb7\xeeme\x06\x17\x85\xe9\xfa\xeb5U\x9c\xcd\x9f\x85\xc9g\x1eF64A\xa9yH\x1f\xacL\xdb\x10\x84}'T\xd3\x00xW\x99`\xac\xb3\xbb\x9a\x973E\xfc\x11j\x98\x9b\xab\xb7Z\x01j\xca_\xd4N(gJ\xa6\x93q>\xb7\xf0\x8a1\xe6\x86\x8f=\xcb\xf6\x8a\x13\xceJe\xb4\xc5\x87w-\xce:\xb1\nq\x01\xd1\xed\x8aoG\xda(UB\x17i\x14(\xf1\xb3-\xc9 \xdb\xf7-152\xd9#\x8d\xa29\xb7\xea\x89\xce]\x05j\x9b\x97\x08\x87$\x9fB Y\x07\x0fw\x14\xd6\xaa\x95\n\xf9V\x84s\xe7\xbcm\xe5\xf1\xdd\xd2\xba\xaa\xf2<{\xc7\xa4\xddJ\x89\xbb\xeb@gp\x081\xc5`\x8d\xf5\x8eP\xb4\x17F/\xbb\xd6\xa6\xd7\xe7\xc7\xfb\xa2\x9d\x80\xd69Q\xc1d\x88\x9b\x11\x96\x91\x1a",
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "BookSummary",
                "description": None,
                "schema": {
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"title": "Author", "type": "string"},
                        "pages": {"title": "Pages", "type": "integer"},
                        "publication_year": {
                            "title": "Publication Year",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "pages", "publication_year"],
                    "title": "BookSummary",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [
                {
                    "name": "get_book_info",
                    "description": "Look up book information by ISBN.",
                    "parameters": """\
{
  "properties": {
    "isbn": {
      "title": "Isbn",
      "type": "string"
    }
  },
  "required": [
    "isbn"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
            "n_chunks": 3,
        }
    }
)



import inspect

import pytest

from mirascope.llm.responses._utils import extract_serialized_json


def test_extract_json_from_code_block() -> None:
    text = inspect.cleandoc("""
        ```json
        {"key": "value"}
        ```
    """)
    result = extract_serialized_json(text)
    assert result == '{"key": "value"}'


def test_extract_json_from_code_block_json_contains_code_block() -> None:
    text = inspect.cleandoc("""
        ```json
        {"key": "value": "```json\n{"is_cursed_object": true}\n```"}
        ```
    """)
    result = extract_serialized_json(text)
    assert result == '{"key": "value": "```json\n{"is_cursed_object": true}\n```"}'


def test_extract_json_from_code_block_with_escaped_quotes() -> None:
    text = inspect.cleandoc("""
        ```json
        {"message": "He said \"hello\" and she replied \"```json\n{\"nested\": true}\n```\""}
        ```
    """)
    result = extract_serialized_json(text)
    assert (
        result
        == '{"message": "He said "hello" and she replied "```json\n{"nested": true}\n```""}'
    )


def test_extract_json_from_code_block_with_prefix_text() -> None:
    text = inspect.cleandoc("""
        Sure thing! Here's the JSON:
        ```json
        {"key": "value"}
        ```
    """)
    result = extract_serialized_json(text)
    assert result == '{"key": "value"}'


def test_extract_json_from_multiple_code_blocks() -> None:
    text = inspect.cleandoc("""
        Sure thing! Here's the JSON:
        ```json
        {"key": "value"}
        ```

        Here's some more!!!
        ```json
        {"key2": "value2"}
        ```
    """)
    result = extract_serialized_json(text)
    assert result == '{"key": "value"}'


def test_code_block_takes_precedence_over_bare_json() -> None:
    text = inspect.cleandoc("""
        Sure thing! I'll use that {} (json) notation you mentioned:
        ```json
        {"key": "value"}
        ```
    """)
    result = extract_serialized_json(text)
    assert result == '{"key": "value"}'


def test_extract_json_from_bare_json() -> None:
    text = '{"key": "value"}'
    result = extract_serialized_json(text)
    assert result == '{"key": "value"}'


def test_extract_json_from_bare_json_with_prefix_and_suffix() -> None:
    text = inspect.cleandoc("""
        The result is: {"key": "value"} and that's final
    """)
    result = extract_serialized_json(text)
    assert result == '{"key": "value"}'


def test_extract_nested_json() -> None:
    text = '{"outer": {"inner": {"key": "value"}}}'
    result = extract_serialized_json(text)
    assert result == '{"outer": {"inner": {"key": "value"}}}'


def test_unfinished_json_code_block() -> None:
    text = inspect.cleandoc("""
        ```json
        {"key": "value"}
    """)
    result = extract_serialized_json(text)
    assert result == '{"key": "value"}'


def test_no_opening_brace_raises_error() -> None:
    text = inspect.cleandoc("""
        This has no JSON
    """)
    with pytest.raises(ValueError, match="no opening `{`"):
        extract_serialized_json(text)


def test_no_closing_brace_raises_error() -> None:
    text = inspect.cleandoc("""
        {"key": "value"
    """)
    with pytest.raises(ValueError, match="no closing `}`"):
        extract_serialized_json(text)


def test_empty_string_raises_error() -> None:
    text = ""
    with pytest.raises(ValueError, match="no opening `{`"):
        extract_serialized_json(text)


def test_only_braces_returns_empty_object() -> None:
    text = "{}"
    result = extract_serialized_json(text)
    assert result == "{}"


def test_multiple_json_objects_returns_first_to_last() -> None:
    text = '{"first": "object"} some text {"second": "object"}'
    result = extract_serialized_json(text)
    assert result == '{"first": "object"}'


def test_json_with_newlines_and_formatting() -> None:
    text = inspect.cleandoc("""Here's the formatted JSON:
    {
        "key": "value",
        "nested": {
            "inner": "data"
        }
    }
    That's the result.""")
    result = extract_serialized_json(text)
    expected = inspect.cleandoc("""{
        "key": "value",
        "nested": {
            "inner": "data"
        }
    }""")
    assert result == expected


def test_json_code_block_with_multiple_backticks() -> None:
    text = 'Some text ```json\n{"key": "value"}\n``` more text ```\nother content\n```'
    result = extract_serialized_json(text)
    assert result == '{"key": "value"}'


def test_json_with_escaped_braces_in_string() -> None:
    text = '{"message": "This has \\{ and \\} in it"}'
    result = extract_serialized_json(text)
    assert result == '{"message": "This has \\{ and \\} in it"}'


def test_json_with_json_in_string() -> None:
    text = '{"message": "This has {"nested": true} in it"}'
    result = extract_serialized_json(text)
    assert result == '{"message": "This has {"nested": true} in it"}'



"""Tests related to handling of base params."""

from typing import get_type_hints

import pytest

from mirascope import llm


def test_ensure_all_params_accessed() -> None:
    """Verify that ensure_all_params_accessed raises AssertionError when any param is not accessed.

    For each parameter in Params, create a params dict, access all parameters except one,
    and verify that exiting the context manager raises a AssertionError.
    """
    params_keys = set(get_type_hints(llm.Params).keys())
    params_keys.discard("temperature")  # Simulate one parameter being unsupported

    # Confirm failure if we skip any param
    for skipped_param in params_keys:
        with (
            pytest.raises(
                AssertionError,
                match="Mismatch between unsupported and unaccessed params",
            ),
            llm.clients.base._utils.ensure_all_params_accessed(
                params={}, provider="google", unsupported_params=["temperature"]
            ) as param_accessor,
        ):
            for param in params_keys:
                if param != skipped_param:
                    getattr(param_accessor, param)

    # Confirm success if we access every param
    with llm.clients.base._utils.ensure_all_params_accessed(
        params={}, provider="google", unsupported_params=["temperature"]
    ) as param_accessor:
        for param in params_keys:
            getattr(param_accessor, param)


def test_error_if_checking_unsupported_param() -> None:
    """Verify that ensure_all_params_accessed raises AssertionError if an unsupported param gets accessed."""
    params_keys = set(get_type_hints(llm.Params).keys())

    with (
        pytest.raises(
            AssertionError, match="Mismatch between unsupported and unaccessed params"
        ),
        llm.clients.base._utils.ensure_all_params_accessed(
            params={}, provider="google", unsupported_params=["temperature"]
        ) as param_accessor,
    ):
        for param in params_keys:
            getattr(param_accessor, param)



"""The `ToolSchema` class for defining tools that LLMs can request be called."""

from __future__ import annotations

import inspect
from collections import namedtuple
from dataclasses import dataclass
from typing import (
    Annotated,
    Any,
    Generic,
    TypeVar,
    get_args,
    get_origin,
    get_type_hints,
)

from docstring_parser import parse
from pydantic import BaseModel, Field, create_model
from pydantic.fields import FieldInfo

from ..content import ToolCall
from .protocols import AsyncContextToolFn, AsyncToolFn, ContextToolFn, ToolFn

ToolFnT = TypeVar(
    "ToolFnT",
    bound=ToolFn | AsyncToolFn | ContextToolFn | AsyncContextToolFn,
    covariant=True,
)
ToolSchemaT = TypeVar("ToolSchemaT", bound="ToolSchema")

DocstringArg = namedtuple("DocstringArg", ["name", "description"])

FORMAT_TOOL_NAME = "__mirascope_formatted_output_tool__"
"""Reserved name of the formatted output tool. 

Any call to a tool with this name is NOT considered a regular tool call, but will instead
be converted into textual output containing the arguments to the tool call.
"""
# TODO: Investigate whether using this hardcoded tool name has any adverse impact
# on model performance, compared to a tool name that references the name of the formatted
# class.


@dataclass
class ParsedDocstring:
    args: list[DocstringArg]


def _parse_docstring_params(docstring: str | None) -> ParsedDocstring:
    """Parse parameter descriptions from a docstring.

    Uses docstring-parser library which supports ReST, Google, Numpydoc-style and
    Epydoc docstrings

    Args:
        docstring: The function's docstring

    Returns:
        ParsedDocstring containing parameter descriptions
    """
    if not docstring:
        return ParsedDocstring(args=[])

    parsed = parse(docstring)
    args = []

    for param in parsed.params:
        if param.description:
            args.append(
                DocstringArg(name=param.arg_name, description=param.description)
            )

    return ParsedDocstring(args=args)


class ToolParameterSchema(BaseModel):
    """JSON Schema for tool parameters (always an object with properties).

    This contains real JSON Schema as generated by Pydantic, with full support
    for complex schemas like anyOf, nested objects, validation constraints, etc.
    Including $defs for complex type references.
    """

    properties: dict[str, dict[str, Any]] = Field(default_factory=dict)
    """Dictionary mapping parameter names to their JSON Schema definitions."""

    required: list[str] = Field(default_factory=list)
    """List of required parameter names."""

    additionalProperties: bool = False
    """Whether additional properties beyond those defined are allowed."""

    defs: dict[str, dict[str, Any]] | None = Field(default=None, alias="$defs")
    """JSON Schema definitions for complex types referenced via $ref."""


@dataclass
class ToolSchema(Generic[ToolFnT]):
    """Underlying schema defining a tool that can be used by LLMs.

    A ToolSchema represents a function that can be called by an LLM during a call.
    It includes metadata like name, description, and parameter schema.

    This class is not instantiated directly but created by the `@tool()` decorator.
    """

    fn: ToolFnT
    """The function that implements the tool's functionality."""

    name: str
    """The name of the tool, used by the LLM to identify which tool to call."""

    description: str
    """Description of what the tool does, extracted from the function's docstring."""

    parameters: ToolParameterSchema
    """JSON Schema describing the parameters accepted by the tool.
    
    The serialized parameters table is cached for efficient hash lookups (e.g. when
    caching provider-encoded tool representations in a LRU cache). Therefore,
    it should **not be modified** after the ToolSchema is created.
    """

    strict: bool
    """Whether the tool should use strict mode when supported by the model."""

    def __hash__(self) -> int:
        if not hasattr(self, "_hash"):
            self._hash = hash(
                (
                    self.name,
                    self.description,
                    self.strict,
                    self.parameters.model_dump_json(),
                )
            )
        return self._hash

    def __init__(
        self,
        fn: ToolFnT,
        *,
        strict: bool = False,
        is_context_tool: bool = False,
    ) -> None:
        """Create a `ToolSchema` by inspecting a function and its docstring.

        Uses Pydantic's create_model to dynamically build a model from the function
        signature, then extracts the JSON schema from it. This leverages Pydantic's
        robust type conversion and union handling.

        Args:
            fn: The function to extract schema from
            strict: Whether the tool should use strict mode when supported
            is_context_tool: Whether this is a context tool (skips the context parameter)

        Returns:
            a `ToolSchema` representing the function

        Raises:
            ValueError: If the tool has a reserved name.
        """
        name = fn.__name__
        if name.startswith(FORMAT_TOOL_NAME):
            raise ValueError(
                f"Cannot use reserved name {FORMAT_TOOL_NAME} as tool name."
            )
        description = inspect.cleandoc(fn.__doc__) if fn.__doc__ else name

        param_descriptions = _parse_docstring_params(fn.__doc__)

        field_definitions = {}
        hints = get_type_hints(fn, include_extras=True)

        context_param_skipped = False
        for param in inspect.signature(fn).parameters.values():
            # Skip args that are provided by Python or Mirascope, not LLM-generated.
            # TODO: Handling of FromCallArgs
            if param.name in ("self", "cls"):
                continue

            # For context tools, skip the first non-self/cls parameter (the context parameter)
            if is_context_tool and not context_param_skipped:
                context_param_skipped = True
                continue

            param_type = hints.get(param.name, Any)
            default = ... if param.default is inspect.Parameter.empty else param.default
            field_info = None

            if get_origin(param_type) is Annotated:
                args = get_args(param_type)
                param_type = args[0]
                for annotation in args[1:]:
                    if isinstance(annotation, FieldInfo):
                        field_info = annotation
                        break

            if field_info is not None:
                field_value = Field(
                    default=default,
                    description=field_info.description,
                )
            else:
                docstring_description = None
                for arg in param_descriptions.args:
                    if arg.name == param.name:
                        docstring_description = arg.description
                        break
                if docstring_description:
                    field_value = Field(
                        default=default, description=docstring_description
                    )
                else:
                    field_value = default

            field_definitions[param.name] = (param_type, field_value)

        TempModel = create_model("TempModel", **field_definitions)

        schema = TempModel.model_json_schema()

        parameters = ToolParameterSchema(
            properties=schema.get("properties", {}),
            required=schema.get("required", []),
            additionalProperties=False,
        )
        if "$defs" in schema:
            parameters.defs = schema["$defs"]

        self.fn = fn
        self.name = name
        self.description = description
        self.parameters = parameters
        self.strict = strict

    def can_execute(self, tool_call: ToolCall) -> bool:
        """Check if a `ToolCall` can be executed by tools with this `ToolSchema`.

        This method is a convenient way to determine if a `ToolCall` is likely intended
        to be executed by a tool with this `ToolSchema`. It does so by checking
        whether the name on the call matches the name on the schema. No other validation
        is performed.
        """
        return tool_call.name == self.name



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    Thought,
    ToolCall,
    ToolOutput,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider": "openai:responses",
            "model_id": "gpt-4o",
            "params": {"thinking": False},
            "finish_reason": None,
            "messages": [
                UserMessage(content=[Text(text="What is the 100th fibonacci number?")]),
                AssistantMessage(
                    content=[
                        Thought(
                            thought='The user is asking for the 100th Fibonacci number. I have a function available called "compute_fib" that can compute the nth Fibonacci number (1-indexed). The user has provided the specific value n=100, so I have all the required parameters to make the function call.'
                        ),
                        ToolCall(
                            id="toolu_01LVmhRSWyUc2ep2hbP7hH84",
                            name="compute_fib",
                            args='{"n": 100}',
                        ),
                    ],
                    provider="anthropic",
                    model_id="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "signature": "ErQDCkYICBgCKkCsItOUHNinar+NABlYn+UTCEV462H15XiUdax8L7lhYMAvN2X2YxzcMd8Obghv+2Menhq/JUHTvDvU3gt3TiOQEgwBwO65BrVeQDsdzOsaDJLQlWqNjpXdBBjgGiIwd7VNj7mVC11HNOFLqTxqq/O2ERu9FYdGiOFTwMyutumJ2sWNTrrAIzNf9SpAk5SSKpsC0hXQacdwPhqJQxZc8hauodhKwKfpksYOm0yJd7v1aRdrAZHrYwfMYlyiygzctAx3wLXzFE7Ez1frJkF+PZYok76DU7BMP2wC+JSiGsPdeaNcy7RnsP7S2/U86non+98+/LLM6TtJ0tTYl1e5q2H0KFl/J3i+ivRGTkbw7+Nf/LKGxT+luaiZPxwvcSqoQiuRUAx0TV875THH5FWYjxxtRpOFmoYmNglLJ9jkoXc2Z4Tvi1ESojBFWW0aBm8yWPaEb/pnDVKebCXBro+J6Gndpx17Dn74eOv9lmAQqu0QM+ymgMWBeJezBvf7QUQGZ511Dnyliy3HkcahbdT1XMhim6i6lf9Zugcha1lL2ZFHuzaZnMhnG9NxC867vhgB",
                                "thinking": 'The user is asking for the 100th Fibonacci number. I have a function available called "compute_fib" that can compute the nth Fibonacci number (1-indexed). The user has provided the specific value n=100, so I have all the required parameters to make the function call.',
                                "type": "thinking",
                            },
                            {
                                "id": "toolu_01LVmhRSWyUc2ep2hbP7hH84",
                                "input": {"n": 100},
                                "name": "compute_fib",
                                "type": "tool_use",
                            },
                        ],
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="toolu_01LVmhRSWyUc2ep2hbP7hH84",
                            name="compute_fib",
                            value="218922995834555169026",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="The 100th Fibonacci number is 218,922,995,834,555,169,026."
                        )
                    ],
                    provider="openai:responses",
                    model_id="gpt-4o",
                    raw_message=[
                        {
                            "id": "msg_094f26de235459b00068f9669723508194938b9bf3722348d9",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": "The 100th Fibonacci number is 218,922,995,834,555,169,026.",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        }
                    ],
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "compute_fib",
                    "description": "Compute the nth Fibonacci number (1-indexed).",
                    "parameters": """\
{
  "properties": {
    "n": {
      "title": "N",
      "type": "integer"
    }
  },
  "required": [
    "n"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
        }
    }
)



"""The `llm.agent` decorator for turning a function into an agent."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Protocol, overload
from typing_extensions import TypeVar, Unpack

from ..tools import AsyncContextTool, AsyncTool, ContextTool, Tool
from .agent_template import AgentTemplate, AsyncAgentTemplate

if TYPE_CHECKING:
    from ..clients import (
        AnthropicClient,
        AnthropicModelId,
        BaseClient,
        GoogleClient,
        GoogleModelId,
        ModelId,
        OpenAICompletionsClient,
        OpenAICompletionsModelId,
        Params,
        Provider,
    )

from ..context import Context, DepsT
from ..formatting import FormattableT
from ..types import P

AgentToolT = TypeVar(
    "AgentToolT",
    bound="Tool | AsyncTool | ContextTool[Any] | AsyncContextTool[Any] | None",
    covariant=True,
    default=None,
)


class SystemPrompt(Protocol[P]):
    """Protocol for a prompt template function that returns a system prompt as a string (no context)."""

    def __call__(self) -> str: ...


class ContextSystemPrompt(Protocol[P, DepsT]):
    """Protocol for a prompt template function that returns a system prompt as a string (with context)."""

    def __call__(self, ctx: Context[DepsT]) -> str: ...


class AsyncSystemPrompt(Protocol[P]):
    """Protocol for an async prompt template function that returns a system prompt as a string (no context)."""

    async def __call__(self) -> str: ...


class AsyncContextSystemPrompt(Protocol[P, DepsT]):
    """Protocol for an async prompt template function that returns a system prompt as a string (with context)."""

    async def __call__(self, ctx: Context[DepsT]) -> str: ...


class AgentDecorator(Protocol[P, AgentToolT, FormattableT]):
    """Protocol for the `agent` decorator."""

    @overload
    def __call__(
        self: AgentDecorator[
            P,
            None | Tool | ContextTool[DepsT],
            FormattableT,
        ],
        fn: SystemPrompt[P] | ContextSystemPrompt[P, DepsT],
    ) -> AgentTemplate[DepsT, FormattableT]:
        """Decorator for creating a sync agent."""
        ...

    @overload
    def __call__(
        self: AgentDecorator[
            P,
            None | AsyncTool | AsyncContextTool[DepsT],
            FormattableT,
        ],
        fn: AsyncSystemPrompt[P] | AsyncContextSystemPrompt[P, DepsT],
    ) -> AsyncAgentTemplate[DepsT, FormattableT]:
        """Decorator for creating an async agent."""
        ...

    def __call__(
        self,
        fn: SystemPrompt[P]
        | ContextSystemPrompt[P, DepsT]
        | AsyncSystemPrompt[P]
        | AsyncContextSystemPrompt[P, DepsT],
    ) -> AgentTemplate[DepsT, FormattableT] | AsyncAgentTemplate[DepsT, FormattableT]:
        """Decorator for creating an agent."""
        raise NotImplementedError()


@overload
def agent(
    *,
    provider: Literal["anthropic"],
    model_id: AnthropicModelId,
    tools: list[AgentToolT] | None = None,
    format: type[FormattableT] | None = None,
    client: AnthropicClient | None = None,
    **params: Unpack[Params],
) -> AgentDecorator[..., AgentToolT, FormattableT]:
    """Decorator for creating an Anthropic agent."""
    ...


@overload
def agent(
    *,
    provider: Literal["google"],
    model_id: GoogleModelId,
    tools: list[AgentToolT] | None = None,
    format: type[FormattableT] | None = None,
    client: GoogleClient | None = None,
    **params: Unpack[Params],
) -> AgentDecorator[..., AgentToolT, FormattableT]:
    """Decorator for creating a Google agent."""
    ...


@overload
def agent(
    *,
    provider: Literal["openai:completions"],
    model_id: OpenAICompletionsModelId,
    tools: list[AgentToolT] | None = None,
    format: type[FormattableT] | None = None,
    client: OpenAICompletionsClient | None = None,
    **params: Unpack[Params],
) -> AgentDecorator[..., AgentToolT, FormattableT]:
    """Decorator for creating an OpenAI agent."""
    ...


@overload
def agent(
    *,
    provider: Provider,
    model_id: ModelId,
    tools: list[AgentToolT] | None = None,
    format: type[FormattableT] | None = None,
    client: None = None,
    **params: Unpack[Params],
) -> AgentDecorator[..., AgentToolT, FormattableT]:
    """Decorator for creating an agent using any registered model."""
    ...


def agent(
    *,
    provider: Provider,
    model_id: ModelId,
    tools: list[AgentToolT] | None = None,
    format: type[FormattableT] | None = None,
    client: BaseClient | None = None,
    **params: Unpack[Params],
) -> AgentDecorator[..., AgentToolT, FormattableT]:
    """Decorator for creating an agent or structured agent.

    Args:
        model_id: The model to use for the agent.
        tools: The tools available to the agent.
        format: The response format type for the agent.
        client: The client to use for the agent.
        **params: Additional parameters for the model.

    Returns:
        An of `AgentDecorator`.
    """
    raise NotImplementedError()



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    Thought,
    ToolCall,
    ToolOutput,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider": "anthropic",
            "model_id": "claude-sonnet-4-0",
            "params": {"thinking": False},
            "finish_reason": None,
            "messages": [
                UserMessage(content=[Text(text="What is the 100th fibonacci number?")]),
                AssistantMessage(
                    content=[
                        Thought(
                            thought="""\
**Let's Get This Fibonacci Number**

Okay, so I see the user wants the 100th Fibonacci number.  No problem. I immediately think of that handy `compute_fib` function in the `default_api`.  Seems like the perfect tool for this task.  I just need to call `compute_fib` with `n=100`, and I should have the answer in no time. Simple and straightforward, exactly how I like it.
"""
                        ),
                        ToolCall(
                            id="google_unknown_tool_id",
                            name="compute_fib",
                            args='{"n": 100}',
                        ),
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "video_metadata": None,
                                "thought": True,
                                "inline_data": None,
                                "file_data": None,
                                "thought_signature": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "function_call": None,
                                "function_response": None,
                                "text": """\
**Let's Get This Fibonacci Number**

Okay, so I see the user wants the 100th Fibonacci number.  No problem. I immediately think of that handy `compute_fib` function in the `default_api`.  Seems like the perfect tool for this task.  I just need to call `compute_fib` with `n=100`, and I should have the answer in no time. Simple and straightforward, exactly how I like it.
""",
                            },
                            {
                                "video_metadata": None,
                                "thought": None,
                                "inline_data": None,
                                "file_data": None,
                                "thought_signature": b"\n\xcc\x01\x01\xd1\xed\x8aoC>s\xdd\x7f\xa6\xeb\x01\xdf>R\x07\xddI\xc3\xe6x\xbb\xa2\xbd/\x8f\x1c\x05'\x12\x945\xecv6w<\x86\x9d\x89\xb6\xbc\xfd\x85m\xd9\xfb&\xa5MR\xe7]_\r\x16\x1c\x13\x18\xac\xe7\x95\xfcW\x90\x9e\x14O\xf1D\x97\xcb\x87e\xb9\x94R\x8d\xb0\xc4\xed\xf37#\xbb1\xdf\xae\xd7\xf0\xe39\xf3T\x01]\xfb:Q\xae@\xfc\x12;\xfdFj\x02!\x89\x02\x9ez>\xfaK\x940}f\xe2\xdav\xac\x1f\x14l\xffJ\xf0\x0c\xcb?\xdf&\xfb9'z|\xa4\x1a\xc0\x03OQ\x80\xcd\xd0\xd6\x92\xc9hi{f\x07\x1c\xf5X\x0e\x93\x8d\xdf\xad\xb5l]\x8f\x856\xb8V\xa3]!\x8a\x1d\x9b\xd7\xef\xc9Y)\xb6\x01}!\xd7\xb8]\xf8\xdf\xa7\x1bJ\x870\x81N&]@",
                                "code_execution_result": None,
                                "executable_code": None,
                                "function_call": {
                                    "id": None,
                                    "args": {"n": 100},
                                    "name": "compute_fib",
                                },
                                "function_response": None,
                                "text": None,
                            },
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="google_unknown_tool_id",
                            name="compute_fib",
                            value="218922995834555169026",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
The 100th Fibonacci number is **218,922,995,834,555,169,026**.

This is quite a large number! The Fibonacci sequence grows exponentially, so by the time we reach the 100th term, we're dealing with a 21-digit number.\
"""
                        )
                    ],
                    provider="anthropic",
                    model_id="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "citations": None,
                                "text": """\
The 100th Fibonacci number is **218,922,995,834,555,169,026**.

This is quite a large number! The Fibonacci sequence grows exponentially, so by the time we reach the 100th term, we're dealing with a 21-digit number.\
""",
                                "type": "text",
                            }
                        ],
                    },
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "compute_fib",
                    "description": "Compute the nth Fibonacci number (1-indexed).",
                    "parameters": """\
{
  "properties": {
    "n": {
      "title": "N",
      "type": "integer"
    }
  },
  "required": [
    "n"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
        }
    }
)



"""Tests for the call decorator function."""

import pytest
from inline_snapshot import snapshot
from pydantic import BaseModel

from mirascope import llm


class Format(BaseModel):
    """Test format for structured responses."""

    value: str


@pytest.fixture
def tools() -> list[llm.Tool]:
    """Create a mock tool for testing."""

    @llm.tool
    def tool() -> int:
        return 42

    return [tool]


@pytest.fixture
def context_tools() -> list[llm.ContextTool]:
    """Create a mock context tool for testing."""

    @llm.tool
    def context_tool(ctx: llm.Context[int]) -> int:
        return ctx.deps

    return [context_tool]


@pytest.fixture
def async_tools() -> list[llm.AsyncTool]:
    """Create a mock tool for testing."""

    @llm.tool
    async def tool() -> int:
        return 42

    return [tool]


@pytest.fixture
def async_context_tools() -> list[llm.AsyncContextTool]:
    """Create a mock async context tool for testing."""

    @llm.tool
    async def async_context_tool(ctx: llm.Context[int]) -> int:
        return ctx.deps

    return [async_context_tool]


@pytest.fixture
def params() -> llm.clients.Params:
    return {"temperature": 0.7, "max_tokens": 100}


class TestCall:
    """Tests for regular call decorator (no context dependency)."""

    def test_call_decorator_creation_openai(
        self, tools: list[llm.Tool], params: llm.clients.Params
    ) -> None:
        """Test that call decorator creates CallDecorator with correct parameters for OpenAI."""

        decorator = llm.call(
            provider="openai:completions",
            model_id="gpt-4o-mini",
            tools=tools,
            format=Format,
            **params,
        )

        assert decorator.tools is tools
        assert decorator.format is Format
        assert decorator.model.provider == "openai:completions"
        assert decorator.model.model_id == "gpt-4o-mini"
        assert decorator.model.params == params

    def test_creating_sync_call(
        self, tools: list[llm.Tool], params: llm.clients.Params
    ) -> None:
        """Test that call decorator creates CallDecorator with correct parameters for OpenAI."""

        def prompt() -> str:
            return "Please recommend a fantasy book."

        call = llm.call(
            provider="openai:completions",
            model_id="gpt-4o-mini",
            tools=tools,
            format=Format,
            **params,
        )(prompt)

        assert isinstance(call, llm.calls.Call)

        assert call.model.provider == "openai:completions"
        assert call.model.model_id == "gpt-4o-mini"
        assert call.model.params == params

        assert call.toolkit == llm.Toolkit(tools=tools)
        assert call.format is Format
        assert call.fn() == [llm.messages.user("Please recommend a fantasy book.")]

    @pytest.mark.asyncio
    async def test_creating_async_call(
        self, async_tools: list[llm.AsyncTool], params: llm.clients.Params
    ) -> None:
        """Test that call decorator creates CallDecorator with correct parameters for OpenAI."""

        async def async_prompt() -> str:
            return "Please recommend a fantasy book."

        call = llm.call(
            provider="openai:completions",
            model_id="gpt-4o-mini",
            tools=async_tools,
            format=Format,
            **params,
        )(async_prompt)

        assert isinstance(call, llm.calls.AsyncCall)

        assert call.model.provider == "openai:completions"
        assert call.model.model_id == "gpt-4o-mini"
        assert call.model.params == params

        assert call.toolkit == llm.AsyncToolkit(tools=async_tools)
        assert call.format is Format
        assert await call.fn() == [
            llm.messages.user("Please recommend a fantasy book.")
        ]

    @pytest.mark.vcr()
    def test_call_decorator_e2e_model_override(self) -> None:
        # TODO: Remove e2e tests from non-e2e test directory; either add model
        # override test coverage to e2e, or refactor this to use mocking etc.
        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        def call() -> str:
            return "What company created you? Answer in just one word."

        assert call().pretty() == snapshot("OpenAI.")
        with llm.model(provider="google", model_id="gemini-2.0-flash"):
            assert call().pretty() == snapshot("Google.\n")
            with llm.model(provider="anthropic", model_id="claude-sonnet-4-0"):
                assert call().pretty() == snapshot("Anthropic")

    def test_value_error_invalid_provider(self) -> None:
        with pytest.raises(ValueError, match="Unknown provider: nonexistent"):
            llm.call(provider="nonexistent", model_id="gpt-4o-mini")  # pyright: ignore[reportCallIssue, reportArgumentType]


class TestContextCall:
    """Tests for context call decorator (with context dependency)."""

    def test_context_call_decorator_creation_openai(
        self, context_tools: list[llm.ContextTool], params: llm.clients.Params
    ) -> None:
        """Test that context_call decorator creates ContextCallDecorator with correct parameters for OpenAI."""

        decorator = llm.call(
            provider="openai:completions",
            model_id="gpt-4o-mini",
            tools=context_tools,
            format=Format,
            **params,
        )

        assert decorator.tools is context_tools
        assert decorator.format is Format
        assert decorator.model.provider == "openai:completions"
        assert decorator.model.model_id == "gpt-4o-mini"
        assert decorator.model.params == params

    def test_creating_sync_context_call(
        self, context_tools: list[llm.ContextTool], params: llm.clients.Params
    ) -> None:
        """Test that context_call decorator creates ContextCall with correct parameters."""

        def context_prompt(ctx: llm.Context[int]) -> str:
            return f"Please recommend a fantasy book. My context value is {ctx.deps}."

        call = llm.call(
            provider="openai:completions",
            model_id="gpt-4o-mini",
            tools=context_tools,
            format=Format,
            **params,
        )(context_prompt)

        assert isinstance(call, llm.calls.ContextCall)

        assert call.model.provider == "openai:completions"
        assert call.model.model_id == "gpt-4o-mini"
        assert call.model.params == params

        assert call.toolkit == llm.ContextToolkit(tools=context_tools)
        assert call.format is Format
        ctx = llm.Context(deps=42)
        assert call.fn(ctx) == [
            llm.messages.user(
                "Please recommend a fantasy book. My context value is 42."
            )
        ]

    @pytest.mark.asyncio
    async def test_creating_async_context_call(
        self,
        async_context_tools: list[llm.AsyncContextTool],
        params: llm.clients.Params,
    ) -> None:
        """Test that context_call decorator creates AsyncContextCall with correct parameters."""

        async def async_context_prompt(ctx: llm.Context[int]) -> str:
            return f"Please recommend a fantasy book. My context value is {ctx.deps}."

        call = llm.call(
            provider="openai:completions",
            model_id="gpt-4o-mini",
            tools=async_context_tools,
            format=Format,
            **params,
        )(async_context_prompt)

        assert isinstance(call, llm.calls.AsyncContextCall)

        assert call.model.provider == "openai:completions"
        assert call.model.model_id == "gpt-4o-mini"
        assert call.model.params == params

        assert call.toolkit == llm.AsyncContextToolkit(tools=async_context_tools)
        assert call.format is Format
        ctx = llm.Context(deps=42)
        assert await call.fn(ctx) == [
            llm.messages.user(
                "Please recommend a fantasy book. My context value is 42."
            )
        ]

    def test_context_call_decorator_with_mixed_tools(
        self, tools: list[llm.Tool], params: llm.clients.Params
    ) -> None:
        """Test that context_call decorator works with regular tools too."""

        def context_prompt(ctx: llm.Context[int]) -> str:
            return f"Please recommend a fantasy book. My context value is {ctx.deps}."

        call = llm.call(
            provider="openai:completions",
            model_id="gpt-4o-mini",
            tools=tools,
            format=Format,
            **params,
        )(context_prompt)

        assert isinstance(call, llm.calls.ContextCall)
        assert call.toolkit == llm.ContextToolkit(tools=tools)

    @pytest.mark.vcr()
    def test_context_call_decorator_e2e_model_override(self) -> None:
        # TODO: Remove e2e tests from non-e2e test directory; either add model
        # override test coverage to e2e, or refactor this to use mocking etc.
        ctx = llm.Context(deps="Who created you?")

        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        def call(ctx: llm.Context[str]) -> str:
            return f"Answer the question in just one word: {ctx.deps}."

        assert call(ctx).pretty() == snapshot("OpenAI.")
        with llm.model(provider="google", model_id="gemini-2.0-flash"):
            assert call(ctx).pretty() == snapshot("Google.\n")
            with llm.model(provider="anthropic", model_id="claude-sonnet-4-0"):
                assert call(ctx).pretty() == snapshot("Anthropic.")

    def test_context_parameter_name_independence(self) -> None:
        """Test that context prompts require the first parameter be named ctx."""

        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        def context_weird_name(special_context: llm.Context[str]) -> str:
            return f"Value: {special_context.deps}"

        assert isinstance(context_weird_name, llm.calls.Call)

    def test_async_context_parameter_name_independence(self) -> None:
        """Test that async context prompts require the first parameter be named ctx."""

        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        async def context_weird_name(special_context: llm.Context[str]) -> str:
            return f"Value: {special_context.deps}"

        assert isinstance(context_weird_name, llm.calls.AsyncCall)

    def test_non_context_typed_parameter(self) -> None:
        """Test that non-Context typed parameters are not treated as context prompts."""

        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        def non_context_prompt(ctx: int) -> str:
            return f"Value: {ctx}"

        assert isinstance(non_context_prompt, llm.calls.Call)

    def test_async_non_context_typed_parameter(self) -> None:
        """Test that non-Context typed async parameters are not treated as context prompts."""

        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        async def non_context_prompt(ctx: int) -> str:
            return f"Value: {ctx}"

        assert isinstance(non_context_prompt, llm.calls.AsyncCall)

    def test_missing_type_annotation(self) -> None:
        """Test that missing type annotations are not treated as context prompts."""

        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        def missing_annotation_prompt(ctx) -> str:  # noqa: ANN001
            return "hi"

        assert isinstance(missing_annotation_prompt, llm.calls.Call)

    def test_async_missing_type_annotation(self) -> None:
        """Test that missing type annotations are not treated as async context prompts."""

        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        async def missing_annotation_prompt(ctx) -> str:  # noqa: ANN001
            return "hi"

        assert isinstance(missing_annotation_prompt, llm.calls.AsyncCall)

    def test_method_with_context_after_self(self) -> None:
        """Test that methods with self as first arg and Context as second arg are context prompts."""

        class TestClass:
            @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
            def method_with_context(self, ctx: llm.Context[str]) -> str:
                return f"Value: {ctx.deps}"

        assert isinstance(TestClass.method_with_context, llm.calls.ContextCall)

    def test_async_method_with_context_after_self(self) -> None:
        """Test that async methods with self as first arg and Context as second arg are context prompts."""

        class TestClass:
            @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
            async def method_with_context(self, ctx: llm.Context[str]) -> str:
                return f"Value: {ctx.deps}"

        assert isinstance(TestClass.method_with_context, llm.calls.AsyncContextCall)

    def test_context_not_first_parameter(self) -> None:
        """Test that Context as second parameter (after non-self) is not treated as context prompt."""

        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        def second_arg_context(regular_param: int, ctx: llm.Context[str]) -> str:
            return f"Value: {regular_param}-{ctx.deps}"

        assert isinstance(second_arg_context, llm.calls.Call)

    def test_async_context_not_first_parameter(self) -> None:
        """Test that Context as second parameter (after non-self) is not treated as async context prompt."""

        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        async def second_arg_context(regular_param: int, ctx: llm.Context[str]) -> str:
            return f"Value: {regular_param}-{ctx.deps}"

        assert isinstance(second_arg_context, llm.calls.AsyncCall)

    def test_context_subclass_detection(self) -> None:
        """Test that Context subclasses are properly detected as context prompts."""

        class CustomContext(llm.Context[str]): ...

        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        def with_custom_context(ctx: CustomContext) -> str:
            return str(ctx.deps)

        assert isinstance(with_custom_context, llm.calls.ContextCall)

    def test_async_context_subclass_detection(self) -> None:
        """Test that Context subclasses are properly detected as async context prompts."""

        class CustomContext(llm.Context[str]): ...

        @llm.call(provider="openai:completions", model_id="gpt-4o-mini")
        async def with_custom_context(ctx: CustomContext) -> str:
            return str(ctx.deps)

        assert isinstance(with_custom_context, llm.calls.AsyncContextCall)



"""End-to-end tests for resuming with override while using thinking and tools."""

from typing import TypedDict

import pytest

from mirascope import llm
from tests.e2e.conftest import PROVIDER_MODEL_ID_PAIRS
from tests.utils import Snapshot, snapshot_test


class ProviderAndModelId(TypedDict, total=True):
    provider: llm.Provider
    model_id: llm.ModelId


def default_provider_and_model(
    provider: llm.Provider,
) -> ProviderAndModelId:
    """Default provider and model that are distinct from the provider being tested.

    Used to ensure that we can test having the provider under test resume
    from a response that was created by a different provider.
    """
    if provider == "anthropic":
        return {"provider": "google", "model_id": "gemini-2.5-flash"}
    else:
        return {"provider": "anthropic", "model_id": "claude-sonnet-4-0"}


@pytest.mark.parametrize("provider,model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
def test_resume_with_override_thinking_and_tools(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test call with thinking and tools, resuming with a different model."""

    @llm.tool
    def compute_fib(n: int) -> str:
        """Compute the nth Fibonacci number (1-indexed)."""
        if n == 100:
            return "218922995834555169026"
        return "Not implemented for other values"

    @llm.call(
        **default_provider_and_model(provider),
        thinking=True,
        tools=[compute_fib],
    )
    def fib_query() -> str:
        return "What is the 100th fibonacci number?"

    with snapshot_test(snapshot) as snap:
        primer_response = fib_query()
        assert len(primer_response.tool_calls) >= 1, (
            f"Expected at least one tool call in first response: {primer_response.pretty()}"
        )

        tool_outputs = primer_response.execute_tools()
        with llm.model(provider=provider, model_id=model_id, thinking=False):
            response = primer_response.resume(tool_outputs)

        snap.set_response(response)



import contextlib
from collections.abc import AsyncIterator, Callable
from datetime import timedelta

import httpx
from mcp import ClientSession
from mcp.client.session import ListRootsFnT, SamplingFnT
from mcp.client.sse import sse_client as mcp_sse_client
from mcp.client.stdio import StdioServerParameters
from mcp.client.streamable_http import (
    streamablehttp_client as mcp_streamablehttp_client,
)
from mcp.shared._httpx_utils import McpHttpClientFactory

from ..tools import AsyncTool


class MCPClient:
    """Mirascope wrapper around a MCP ClientSession.

    It provides a way to get MCP results that are pre-converted into Mirascope-friendly
    types.

    The underlying MCP ClientSession may be accessed by .session if needed.
    """

    def __init__(self, session: ClientSession) -> None:
        self._session = session

    @property
    def session(self) -> ClientSession:
        """Access the underlying MCP ClientSession if needed."""
        return self._session

    async def list_tools(self) -> list[AsyncTool]:
        """List all tools available on the MCP server.

        Returns:
            A list of dynamically created `llm.Tool`s.
        """
        raise NotImplementedError()


@contextlib.asynccontextmanager
async def streamablehttp_client(
    url: str,
    headers: dict[str, str] | None = None,
    timeout: float | timedelta | None = None,
    sse_read_timeout: float | timedelta | None = None,
    terminate_on_close: bool = True,
    httpx_client_factory: McpHttpClientFactory | None = None,
    auth: httpx.Auth | None = None,
) -> AsyncIterator[MCPClient]:
    """Create a Mirascope MCPClient using StreamableHTTP."""
    kwargs = {}
    if headers is not None:
        kwargs["headers"] = headers
    if timeout is not None:
        kwargs["timeout"] = timeout
    if sse_read_timeout is not None:
        kwargs["sse_read_timeout"] = sse_read_timeout
    if httpx_client_factory is not None:
        kwargs["httpx_client_factory"] = httpx_client_factory
    if auth is not None:
        kwargs["auth"] = auth

    async with (
        mcp_streamablehttp_client(
            url,
            terminate_on_close=terminate_on_close,
            **kwargs,
        ) as (read, write, get_session_id),
        ClientSession(
            read,
            write,
        ) as session,
    ):
        await session.initialize()
        yield MCPClient(session)


@contextlib.asynccontextmanager
async def stdio_client(
    server_parameters: StdioServerParameters,
    read_stream_exception_handler: Callable[[Exception], None] | None = None,
) -> AsyncIterator[MCPClient]:
    """Create a Mirascope MCPClient using stdio."""

    async with (
        ClientSession(None, None) as session,  # pyright: ignore [reportArgumentType]
    ):
        raise NotImplementedError()
        await session.initialize()
        yield MCPClient(session)


@contextlib.asynccontextmanager
async def sse_client(
    url: str,
    list_roots_callback: ListRootsFnT | None = None,
    read_timeout_seconds: timedelta | None = None,
    sampling_callback: SamplingFnT | None = None,
    session: ClientSession | None = None,
) -> AsyncIterator[MCPClient]:
    """Create a Mirascope MCPClient using sse."""

    async with (
        mcp_sse_client(url) as (read, write),
        ClientSession(
            read,
            write,
            read_timeout_seconds=read_timeout_seconds,
            sampling_callback=sampling_callback,
            list_roots_callback=list_roots_callback,
        ) as session,
    ):
        await session.initialize()
        yield MCPClient(session)



"""The `Tool` class for defining tools that LLMs can request be called."""

from __future__ import annotations

import json
from collections.abc import Awaitable
from typing import Any, Generic, TypeVar, cast

from ..content import ToolCall, ToolOutput
from ..context import Context, DepsT
from ..types import AnyP, JsonableCovariantT
from .protocols import (
    AsyncContextToolFn,
    AsyncToolFn,
    ContextToolFn,
    ToolFn,
    _AsyncJsonKwargsCallable,
    _AsyncKwargsCallable,
    _ContextKwargsCallable,
    _KwargsCallable,
)
from .tool_schema import ToolSchema

ToolT = TypeVar(
    "ToolT",
    bound="Tool | AsyncTool | ContextTool[Any] | AsyncContextTool[Any]",
    covariant=True,
)


class Tool(
    ToolSchema[ToolFn[AnyP, JsonableCovariantT]], Generic[AnyP, JsonableCovariantT]
):
    """A tool that can be used by LLMs.

    A `Tool` represents a function that can be called by an LLM during a call.
    It includes metadata like name, description, and parameter schema.

    This class is not instantiated directly but created by the `@tool()` decorator.
    """

    def __init__(
        self, fn: ToolFn[AnyP, JsonableCovariantT], *, strict: bool = False
    ) -> None:
        super().__init__(fn, strict=strict, is_context_tool=False)

    def __call__(self, *args: AnyP.args, **kwargs: AnyP.kwargs) -> JsonableCovariantT:
        """Call the underlying function directly."""
        return self.fn(*args, **kwargs)

    def execute(self, tool_call: ToolCall) -> ToolOutput[JsonableCovariantT]:
        """Execute the tool using an LLM-provided `ToolCall`."""
        kwargs_from_json = json.loads(tool_call.args)
        kwargs_callable = cast(_KwargsCallable[JsonableCovariantT], self.fn)
        result = kwargs_callable(**kwargs_from_json)
        return ToolOutput(id=tool_call.id, value=result, name=self.name)


class AsyncTool(
    ToolSchema[AsyncToolFn[AnyP, JsonableCovariantT]],
    Generic[AnyP, JsonableCovariantT],
):
    """An async tool that can be used by LLMs.

    An `AsyncTool` represents an async function that can be called by an LLM during a call.
    It includes metadata like name, description, and parameter schema.

    This class is not instantiated directly but created by the `@tool()` decorator.
    """

    def __init__(
        self, fn: AsyncToolFn[AnyP, JsonableCovariantT], *, strict: bool = False
    ) -> None:
        super().__init__(fn, strict=strict, is_context_tool=False)

    def __call__(
        self, *args: AnyP.args, **kwargs: AnyP.kwargs
    ) -> Awaitable[JsonableCovariantT]:
        """Call the underlying async function directly."""
        return self.fn(*args, **kwargs)

    async def execute(self, tool_call: ToolCall) -> ToolOutput[JsonableCovariantT]:
        """Execute the async tool using an LLM-provided `ToolCall`."""
        kwargs_from_json = json.loads(tool_call.args)
        kwargs_callable = cast(_AsyncKwargsCallable[JsonableCovariantT], self.fn)
        result = await kwargs_callable(**kwargs_from_json)
        return ToolOutput(id=tool_call.id, value=result, name=self.name)


class ContextTool(
    ToolSchema[ContextToolFn[DepsT, AnyP, JsonableCovariantT]],
    Generic[DepsT, AnyP, JsonableCovariantT],
):
    """Protocol defining a tool that can be used by LLMs.

    A `ContextTool` represents a function that can be called by an LLM during a call.
    It includes metadata like name, description, and parameter schema.

    This class is not instantiated directly but created by the `@tool()` decorator.
    """

    def __init__(
        self,
        fn: ContextToolFn[DepsT, AnyP, JsonableCovariantT],
        *,
        strict: bool = False,
    ) -> None:
        super().__init__(fn, strict=strict, is_context_tool=True)

    def __call__(
        self,
        ctx: Context[DepsT],
        *args: AnyP.args,
        **kwargs: AnyP.kwargs,
    ) -> JsonableCovariantT:
        """Call the underlying function directly with context."""
        return self.fn(ctx, *args, **kwargs)

    def execute(
        self, ctx: Context[DepsT], tool_call: ToolCall
    ) -> ToolOutput[JsonableCovariantT]:
        """Execute the context tool using an LLM-provided `ToolCall`."""
        kwargs_from_json = json.loads(tool_call.args)
        kwargs_callable = cast(
            _ContextKwargsCallable[DepsT, JsonableCovariantT], self.fn
        )
        result = kwargs_callable(ctx, **kwargs_from_json)
        return ToolOutput(id=tool_call.id, value=result, name=self.name)


class AsyncContextTool(
    ToolSchema[AsyncContextToolFn[DepsT, AnyP, JsonableCovariantT]],
    Generic[DepsT, AnyP, JsonableCovariantT],
):
    """Protocol defining an async tool that can be used by LLMs with context.

    An `AsyncContextTool` represents an async function that can be called by an LLM during a call.
    It includes metadata like name, description, and parameter schema.

    This class is not instantiated directly but created by the `@tool()` decorator.
    """

    def __init__(
        self,
        fn: AsyncContextToolFn[DepsT, AnyP, JsonableCovariantT],
        *,
        strict: bool = False,
    ) -> None:
        super().__init__(fn, strict=strict, is_context_tool=True)

    def __call__(
        self,
        ctx: Context[DepsT],
        *args: AnyP.args,
        **kwargs: AnyP.kwargs,
    ) -> Awaitable[JsonableCovariantT]:
        """Call the underlying async function directly with context."""
        return self.fn(ctx, *args, **kwargs)

    async def execute(
        self, ctx: Context[DepsT], tool_call: ToolCall
    ) -> ToolOutput[JsonableCovariantT]:
        """Execute the async context tool using an LLM-provided `ToolCall`."""
        kwargs_from_json = json.loads(tool_call.args)
        kwargs_callable = cast(
            _AsyncJsonKwargsCallable[DepsT, JsonableCovariantT], self.fn
        )
        result = await kwargs_callable(ctx, **kwargs_from_json)
        return ToolOutput(id=tool_call.id, value=result, name=self.name)



from dataclasses import dataclass

from mirascope import llm


@dataclass
class Deps:
    bar: int


@dataclass
class OtherDeps:
    foo: str


def prompt() -> str:
    return "hello world"


def context_prompt(ctx: llm.Context[Deps]) -> str:
    return "hello world"


async def async_prompt() -> str:
    return "hello world"


async def async_context_prompt(ctx: llm.Context[Deps]) -> str:
    return "hello world"


def context() -> llm.Context[Deps]:
    return llm.Context(deps=Deps(bar=3))


@llm.tool
def tool() -> int:
    return 41


@llm.tool
def context_tool(ctx: llm.Context[Deps]) -> int:
    return 42


@llm.tool
async def async_tool() -> int:
    return 41


@llm.tool
async def async_context_tool(ctx: llm.Context[Deps]) -> int:
    return 42


@llm.tool
def context_tool_other_deps(ctx: llm.Context[OtherDeps]):
    return 41


@llm.tool
async def async_context_tool_other_deps(ctx: llm.Context[OtherDeps]):
    return 41


def tool_call() -> llm.ToolCall:
    raise NotImplementedError()


def expect_call(x: llm.calls.Call): ...
def expect_async_call(x: llm.calls.AsyncCall): ...
def expect_context_call(x: llm.calls.ContextCall[..., Deps]): ...
def expect_async_context_call(x: llm.calls.AsyncContextCall[..., Deps]): ...


def expect_tool(x: llm.tools.Tool): ...
def expect_async_tool(x: llm.tools.AsyncTool): ...
def expect_context_tool(x: llm.tools.ContextTool[Deps]): ...
def expect_async_context_tool(x: llm.tools.AsyncContextTool[Deps]): ...


def sanity_checks():
    expect_tool(tool)
    expect_context_tool(context_tool)

    expect_async_tool(async_tool)
    expect_async_context_tool(async_context_tool)



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    Thought,
    ToolCall,
    ToolOutput,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider": "openai:completions",
            "model_id": "gpt-4o",
            "params": {"thinking": False},
            "finish_reason": None,
            "messages": [
                UserMessage(content=[Text(text="What is the 100th fibonacci number?")]),
                AssistantMessage(
                    content=[
                        Thought(
                            thought="The user is asking for the 100th Fibonacci number. I have a function available called `compute_fib` that can compute the nth Fibonacci number (1-indexed). The user has provided the specific value n=100, so I have all the required parameters to make the function call."
                        ),
                        Text(text="I'll compute the 100th Fibonacci number for you."),
                        ToolCall(
                            id="toolu_01UvLhE2u2ZGcdihNhXUm1wp",
                            name="compute_fib",
                            args='{"n": 100}',
                        ),
                    ],
                    provider="anthropic",
                    model_id="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "signature": "ErQDCkYICBgCKkAvaj6zwlcCXh989DjfcJKxhnEl5NQA4aOMRq/nZzCYnhPFXVcF6OMlhlPU/gK+hzNjWeY/3Q7HLQvMtqr3wym/Egy6b14LOeSQ1wVvF24aDN+G5tLY1buyBhc4gSIwmOXjvuWtJb+x3LV1DMpXkb1R97E3wRyIs+LTxLJ7p9LAebDnW4dIFtJt5vTUqK40KpsCz+4VXHewhrDmLzXOo858J2FHSzuQprqwULAPba1kasCHTEt1Z5eSc3thRuGJReZNpCU82BFeESHvrKj45BSrHxKUsrb9eU9JBGdSgdz/cgKdj3LvkRIhrAq6NIG83hgG2cK6lF2Ae361w+KeBIenHj589Dw9ufmd6Nj6mTNvN2WBcDbRZnn+X4tAl8xIelZ0sBG7bvuzBxV7iPv7+N2qWbiLm68PH3UaBjD2/SKcPUbHB5sv5N6laxwSM6RIZy/hg33l0eXQMT0jHjHHK4vI0Flum4eJwN/7rOM4DAnJCHux297HwheEVzARd4L5qQfTr1q9kdFDlrE7B+FVhD0NURmBKDpkcQ2cYCgnK+uRz2+LW/IZqPWurdiVPBgB",
                                "thinking": "The user is asking for the 100th Fibonacci number. I have a function available called `compute_fib` that can compute the nth Fibonacci number (1-indexed). The user has provided the specific value n=100, so I have all the required parameters to make the function call.",
                                "type": "thinking",
                            },
                            {
                                "citations": None,
                                "text": "I'll compute the 100th Fibonacci number for you.",
                                "type": "text",
                            },
                            {
                                "id": "toolu_01UvLhE2u2ZGcdihNhXUm1wp",
                                "input": {"n": 100},
                                "name": "compute_fib",
                                "type": "tool_use",
                            },
                        ],
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="toolu_01UvLhE2u2ZGcdihNhXUm1wp",
                            name="compute_fib",
                            value="218922995834555169026",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="The 100th Fibonacci number is 218,922,995,834,555,169,026."
                        )
                    ],
                    provider="openai:completions",
                    model_id="gpt-4o",
                    raw_message={
                        "content": "The 100th Fibonacci number is 218,922,995,834,555,169,026.",
                        "role": "assistant",
                        "annotations": [],
                    },
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "compute_fib",
                    "description": "Compute the nth Fibonacci number (1-indexed).",
                    "parameters": """\
{
  "properties": {
    "n": {
      "title": "N",
      "type": "integer"
    }
  },
  "required": [
    "n"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
        }
    }
)



from mirascope import llm


def recommend_book(genre: str) -> llm.Response:
    # [!code highlight:2]
    model = llm.use_model(provider="openai", model_id="gpt-4o-mini")
    message = llm.messages.user(f"Please recommend a book in {genre}.")
    return model.call(messages=[message])


def main():
    # [!code highlight:2]
    with llm.model(provider="anthropic", model_id="claude-sonnet-4-0", temperature=1):
        response: llm.Response = recommend_book("fantasy")
        print(response.pretty())


main()



"""Tests for llm.clients.GoogleClient."""

from unittest.mock import MagicMock, patch

from mirascope import llm


def test_custom_base_url() -> None:
    """Test that custom base URL is used for API requests."""
    example_url = "https://example.com"

    with patch("mirascope.llm.clients.google.clients.Client") as mock_client_class:
        mock_client_instance = MagicMock()
        mock_client_class.return_value = mock_client_instance

        google_client = llm.client("google", base_url=example_url)

        mock_client_class.assert_called_once()
        call_args = mock_client_class.call_args
        assert call_args.kwargs["http_options"] is not None
        assert call_args.kwargs["http_options"].base_url == example_url

        assert google_client.client is mock_client_instance


def test_context_manager() -> None:
    """Test nested context manager behavior and get_client() integration."""

    global_client = llm.get_client("google")

    client1 = llm.client("google", api_key="key1")
    client2 = llm.client("google", api_key="key2")

    assert llm.get_client("google") is global_client

    with client1 as ctx1:
        assert ctx1 is client1
        assert llm.get_client("google") is client1

        with client2 as ctx2:
            assert ctx2 is client2
            assert llm.get_client("google") is client2

        assert llm.get_client("google") is client1

    assert llm.get_client("google") is global_client


def test_client_caching() -> None:
    """Test that client() returns cached instances for identical parameters."""
    client1 = llm.client(
        "google", api_key="test-key", base_url="https://api.example.com"
    )
    client2 = llm.client(
        "google", api_key="test-key", base_url="https://api.example.com"
    )
    assert client1 is client2

    client3 = llm.client(
        "google", api_key="different-key", base_url="https://api.example.com"
    )
    assert client1 is not client3

    client4 = llm.client(
        "google", api_key="test-key", base_url="https://different.example.com"
    )
    assert client1 is not client4

    client5 = llm.client("google", api_key=None, base_url=None)
    client6 = llm.get_client("google")
    assert client5 is client6
